"""Heatmap plot class."""

from matplotlib.colors import LinearSegmentedColormap
import seaborn as sns
import numpy as np
from rnavigate import plots
from rnavigate import data


class Heatmap(plots.Plot):
    """Create a heatmap plot.

    Parameters
    ----------
    num_samples : int
        Number of samples to plot.
    structure : rnavigate.data.SecondaryStructure or rnavigate.data.PDB
        Structure object.
    **kwargs
        Keyword arguments passed to :class:`rnavigate.plots.Plot`.

    Attributes
    ----------
    structure : rnavigate.data.SecondaryStructure or rnavigate.data.PDB
        Structure object.
    fig : matplotlib.figure.Figure
        Figure object.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    i : int
        Index of current axes object. Increments with each call to
        :meth:`rnavigate.plots.Heatmap.plot_data`.
    """

    def __init__(self, num_samples, structure, **plot_kwargs):
        """Initialize Heatmap object."""
        super().__init__(num_samples, sharey=True, **plot_kwargs)
        self.structure = structure
        for i in range(num_samples):
            ax = self.get_ax(i)
            ax.set_aspect("equal")
            mn_mx = (0.5, structure.length + 0.5)
            ax.set(
                ylim=mn_mx, xlim=mn_mx, xticks=list(range(50, structure.length + 1, 50))
            )

    def set_figure_size(
        self,
        height_ax_rel=None,
        width_ax_rel=None,
        width_ax_in=2,
        height_ax_in=2,
        height_gap_in=1,
        width_gap_in=0.5,
        top_in=1,
        bottom_in=0.5,
        left_in=0.5,
        right_in=0.5,
    ):
        """Set figure size.

        Parameters
        ----------
        height_ax_rel : float, defaults to None
            Height of each axes object relative to y-axis limits.
        width_ax_rel : float, defaults to None
            Width of each axes object relative to x-axis limits.
        width_ax_in : float, defaults to 2
            Width of each axes object in inches.
        height_ax_in : float, defaults to 2
            Height of each axes object in inches.
        height_gap_in : float, defaults to 1
            Height of gap between axes objects in inches.
        width_gap_in : float, defaults to 0.5
            Width of gap between axes objects in inches.
        top_in : float, defaults to 1
            Top margin in inches.
        bottom_in : float, defaults to 0.5
            Bottom margin in inches.
        left_in : float, defaults to 0.5
            Left margin in inches.
        right_in : float, defaults to 0.5
            Right margin in inches.
        """
        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in,
            left_in=left_in,
            right_in=right_in,
        )

    def plot_data(
        self,
        interactions,
        label,
        levels=None,
        regions=None,
        interpolation=None,
        atom="O2'",
        plot_type="heatmap",
        weights=None,
    ):
        """Plot heatmap and contour data on the current (or specified) axes.

        Parameters
        ----------
        interactions : rnavigate.data.Interactions
            Interactions object containing inter-nucleotide data to plot as heatmap.
        label : str
            Label for plot title.
        levels : list of float, defaults to None
            Levels for contour lines. If None, levels are set automatically.
            [5] for SecondaryStructure, [20] for PDB.
        regions : list of tuple of int, defaults to None
            Regions to plot as contour lines. If None, distances are plotted.
        interpolation : str, defaults to None
            Interpolation method for heatmap. If None, no interpolation is used.
        atom : str, defaults to "O2'"
            Atom to use for calculating distances.
        plot_type : "heatmap" or "kde", defaults to "heatmap"
            Type of plot to create.
        weights : array-like, defaults to None
            A column name of `interactions` to use as weights for KDE plot.
        """
        ax = self.get_ax()
        if regions is not None:
            self.plot_contour_regions(ax, interactions, regions)
        else:
            self.plot_contour_distances(ax, levels, atom)
        assert plot_type in ["heatmap", "kde"], "plot_type must be heatmap or kde"
        if plot_type == "heatmap":
            self.plot_heatmap_data(ax, interactions, interpolation)
        elif plot_type == "kde":
            self.plot_kde_data(ax, interactions, weights=weights)
        ax.set_title(label)
        self.i += 1

    def plot_contour_regions(self, ax, interactions, regions):
        """Plot contour regions on the current (or specified) axes.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to plot on.
        interactions : rnavigate.data.Interactions
            Interactions object containing inter-nucleotide data to plot as heatmap.
        regions : list of tuple of int
            Regions to plot as contour lines.
        """
        matrix = np.full([interactions.length, interactions.length], 0)
        for (mn1, mx1), (mn2, mx2) in regions:
            matrix[mn1 - 1 : mx1 - 1, mn2 - 1 : mx2 - 1] += 1
        levels = [0.5]
        cmap = LinearSegmentedColormap.from_list("contours", ["black", "gray"])
        x_y = list(range(1, interactions.length + 1))
        ax.contour(x_y, x_y, matrix, levels=levels, cmap=cmap, linewidths=0.3)

    def plot_contour_distances(self, ax, levels, atom):
        """Plot contour distances on the current (or specified) axes.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to plot on.
        levels : list of float, defaults to None
            Levels for contour lines. If None, levels are set automatically.
            [5] for SecondaryStructure, [20] for PDB.
        atom : str, defaults to "O2'"
            Atom to use for calculating distances.
        """
        structure = self.structure
        distances = structure.get_distance_matrix(atom)
        for i in range(structure.length):
            for j in range(i, structure.length):
                distances[i, j] = 0
        if (levels is None) and isinstance(structure, data.SecondaryStructure):
            levels = [5]
        elif (levels is None) and isinstance(structure, data.PDB):
            levels = [20]
        cmap = LinearSegmentedColormap.from_list("contours", ["black", "gray"])
        x_y = list(range(1, structure.length + 1))
        ax.contour(x_y, x_y, distances, levels=levels, cmap=cmap, linewidths=0.3)

    def plot_heatmap_data(self, ax, interactions, interpolation):
        """Plot heatmap data on the specified axes.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to plot on.
        interactions : rnavigate.data.Interactions
            Interactions object containing inter-nucleotide data to plot as heatmap.
        interpolation : str, defaults to None
            Interpolation method for heatmap. If None, no interpolation is used.
        """
        data = interactions.get_sorted_data()
        metric = interactions.metric
        data_im = np.full([interactions.length] * 2, np.nan)
        window = interactions.window
        for _, row in data.iterrows():
            i = int(row["i"] - 1)
            j = int(row["j"] - 1)
            data_im[j : j + window, i : i + window] = row[metric]
        ax.imshow(
            data_im,
            cmap=interactions.cmap.cmap,
            norm=interactions.cmap.norm,
            interpolation=interpolation,
        )

    def plot_kde_data(self, ax, interactions, weights=None, **kwargs):
        """Plot KDE data on the specified axes.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to plot on.
        interactions : rnavigate.data.Interactions
            Interactions object containing inter-nucleotide data to plot as KDE.
        weights : array-like, defaults to None
            A column name of `interactions` to use as weights for KDE plot.
        **kwargs
            Keyword arguments passed to :func:`seaborn.kdeplot`.
        """
        data = interactions.get_sorted_data()
        sns.kdeplot(
            ax=ax,
            data=data,
            x="i",
            y="j",
            fill=True,
            levels=5,
            bw_adjust=0.2,
            cmap=interactions.cmap.cmap,
            common_norm=True,
            weights=weights,
            **kwargs
        )
        ax.set(xlabel="Nucleotide position", ylabel="Nucleotide position")



"""Contains all rnavigate convenience plotting functions"""

from rnavigate import plots
from rnavigate import data
from rnavigate.helper_functions import (
    PlottingArgumentParser,
    _parse_plot_kwargs,
    fit_data,
)
from rnavigate.data_loading import get_sequence


__all__ = [
    "plot_options",
    "plot_qc",
    "plot_shapemapper",
    "plot_skyline",
    "plot_profile",
    "plot_alignment",
    "plot_arcs",
    "plot_arcs_compare",
    "plot_ss",
    "plot_mol",
    "plot_heatmap",
    "plot_circle",
    "plot_linreg",
    "plot_roc",
    "plot_disthist",
    "plot_ntdist",
]


def plot_options(samples):
    """Prints a list of plotting functions compatible with a sample or list of samples.

    Some plotting functions require specific data classes to be loaded into the
    sample. For plotting multiple samples, data keywords that are not shared, or are
    shared, but are not of the same data class, are considered invalid.

    Parameters
    ----------
        samples : rnavigate.Sample or list of rnavigate.Sample
            samples to check for compatible plotting functions
    """
    if not isinstance(samples, list):
        samples = [samples]
    if len(samples) == 0:
        raise ValueError("No samples provided.")
    # get all data keywords for each sample
    # remove " (default)" from data keyword strings
    # find data keywords that are shared by all samples
    data_keywords = samples[0].print_data_keywords(return_dict=True)
    for key in data_keywords:
        for i, keyword in enumerate(data_keywords[key]):
            keyword = keyword.split(" ")[0]
            data_keywords[key][i] = keyword
    for sample in samples[1:]:
        these_data_keywords = sample.print_data_keywords(return_dict=True)
        for key in data_keywords:
            data_keywords[key] = list(
                set(data_keywords[key]) & set(these_data_keywords[key])
            )
    # SHAPEMaP profiles and those with log data (read length and muts/mol)
    shapemap = []
    shapemap_with_log = []
    for keyword in data_keywords["profiles"]:
        if all(isinstance(s.get_data(keyword), data.SHAPEMaP) for s in samples):
            shapemap.append(keyword)
        if all(s.get_data(keyword).read_lengths is None for s in samples):
            shapemap_with_log.append(keyword)
    # SecondaryStructures which contain drawing coordinates
    ss_with_coords = []
    for keyword in data_keywords["structures"]:
        if all("X_coordinate" in s.get_data(keyword).data.columns for s in samples):
            ss_with_coords.append(keyword)
    # any sequence data
    sequences = []
    for key in data_keywords:
        for keyword in data_keywords[key]:
            if all(isinstance(s.get_data(keyword), data.Sequence) for s in samples):
                sequences.append(keyword)

    print(
        """
All data keywords are optional inputs for each RNAvigate sample. However,
each plotting function has specific data requirements. This list contains
all plotting functions for which requirements are met by these samples,
and the specific data keywords that satisfy each requirement.

Further below, a list of compatible optional data keywords is provided.
These are likely to be useful for plotting, but are not required.


Compatible plotting functions and keywords for these samples:"""
    )

    if len(shapemap_with_log) > 0 and len(samples) == 1:
        print(
            f"""
    plot_qc
        compatible profiles (SHAPE-MaP with log data): {shapemap_with_log}"""
        )
    else:
        print(
            """
    plot_qc
    NOT COMPATIBLE
        accepts only 1 sample which must conatin a SHAPE-MaP profile
        with associated log data."""
        )
    if len(shapemap) > 0 and len(samples) == 1:
        print(
            f"""
    plot_shapemapper
        compatible profiles (SHAPE-MaP): {shapemap}"""
        )
    else:
        print(
            """
    plot_shapemapper
    NOT COMPATIBLE
        accepts only 1 sample which must conatin a SHAPE-MaP profile."""
        )
    if len(data_keywords["profiles"]) > 0:
        print(
            f"""
    plot_skyline
    plot_profile
    plot_ntdist
        compatible profiles: {data_keywords["profiles"]}"""
        )
        if len(samples) > 1:
            print(
                """
    plot_linreg
        compatible profiles: {data_keywords["profiles"]}"""
            )
        else:
            print(
                """
    plot_linreg
    NOT COMPATIBLE
        accepts only 2 or more samples which must contain profiles."""
            )
    else:
        print(
            """
    plot_skyline
    plot_profile
    plot_ntdist
    plot_linreg
    NOT COMPATIBLE
        requires sample(s) which contain profiles."""
        )
    if len(sequences) > 0:
        print(
            f"""
    plot_arcs
    plot_circle
        sequences: {sequences}"""
        )
        if len(samples) == 2:
            print(
                f"""
    plot_alignment
    plot_arcs_compare
        sequences: {sequences}"""
            )
        else:
            print(
                """
    plot_alignment
    plot_arcs_compare
    NOT COMPATIBLE
        accepts only 2 samples which must contain sequences."""
            )
    else:
        print(
            """
    plot_arcs
    plot_circle
    plot_alignment
    plot_arcs_compare
    NOT COMPATIBLE
        requires sample(s) which contain sequences."""
        )
    if len(ss_with_coords) > 0:
        print(
            f"""
    plot_ss
        structures: {ss_with_coords}"""
        )
    else:
        print(
            """
    plot_ss
    NOT COMPATIBLE
        requires sample(s) which contain structures with coordinates."""
        )
    if len(data_keywords["pdbs"]) > 0:
        print(
            f"""
    plot_mol
        tertiary structures: {data_keywords["pdbs"]}"""
        )
    else:
        print(
            """
    plot_mol
    NOT COMPATIBLE
        requires sample(s) which contain tertiary structures."""
        )
    if len(data_keywords["interactions"]) > 0:
        print(
            f"""
    plot_heatmap
        interactions: {data_keywords["interactions"]}"""
        )
    else:
        print(
            """
    plot_heatmap
    NOT COMPATIBLE
        requires sample(s) which contain interactions."""
        )
    if len(data_keywords["profiles"]) > 0 and len(data_keywords["structures"]) > 0:
        print(
            f"""
    plot_roc
        profiles: {data_keywords["profiles"]}
        structures: {data_keywords["structures"]}"""
        )
    else:
        print(
            """
    plot_roc
    NOT COMPATIBLE
        requires sample(s) which contain profiles and structures."""
        )
    # requires interactions and secondary or tertiary structure data
    if len(data_keywords["interactions"]) > 0 and (
        len(data_keywords["structures"]) > 0 or len(data_keywords["pdbs"]) > 0
    ):
        print(
            f"""
    plot_disthist
        interactions: {data_keywords["interactions"]}
        structures: {data_keywords["structures"]}
        tertiary structures: {data_keywords["pdbs"]}"""
        )
    else:
        print(
            """
    plot_disthist
    NOT COMPATIBLE
        requires sample(s) which contain interactions and structures."""
        )

    print("\n\nCompatible optional data for plotting functions\n")
    for key in data_keywords:
        print(f"    {key}: {data_keywords[key]}")


def plot_qc(
    # required
    samples,
    profile,
    # optional display
    labels=None,
):
    """Creates a multipanel quality control plot displaying mutations per
    molecule, read length distribution, and mutation rate distributions for
    modified and unmodified samples.

    Parameters
    ----------
        samples : list of rnavigate.Sample
            samples to retrieve data from
        profile : data keyword string or data object
            ShapeMaP or similar data for plotting reactivity distributions
            Must contain data from ShapeMapper log file
        labels : list of str, defaults to sample.sample for each sample in `samples`
            labels to be used on legends, must be same length as samples list

    Returns
    -------
        rnavigate.plots.QC
            the quality control plot object
    """
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        profile=profile,
    )
    plot = plots.QC(num_samples=len(samples))
    for data_dict in parsed_args.data_dicts:
        plot.plot_data(**data_dict)
    plot.set_figure_size()
    return plot


def plot_shapemapper(
    # required
    sample,
    profile,
    # optional display
    label=None,
    panels=None,
):
    """Makes a standard ShapeMapper2 profile plot with 3 panels: Normalized
    Reactivities, mutation rates, and read depths.

    Parameters
    ----------
        sample : rnavigate Sample
            The sample from which data profile and label will be retreived
        profile : data keyword string or data object
            ShapeMaP or similar data for plotting profiles
        label : str, defaults to sample.sample
            A label to use as the title of the figure
        panels : list of str, defaults to ["profile", "rates", "depth"]
            Which panels to include: options are "profile", "rates", and "depth"

    Returns
    -------
        rnavigate.plots.SM
            the ShapeMapper2 plot object
    """
    if panels is None:
        panels = ["profile", "rates", "depth"]
    if label is None:
        label = sample.sample
    profile = sample.get_data(profile)
    plot = plots.SM(profile.length, panels=panels)
    plot.plot_data(profile=profile, label=label)
    plot.set_figure_size()
    return plot


def plot_skyline(
    # required
    samples,
    profile,
    # optional data inputs
    sequence=None,
    annotations=None,
    domains=None,
    # optional data display
    labels=None,
    nt_ticks=(20, 5),
    columns=None,
    errors=None,
    annotations_mode="track",
    seqbar=True,
    region="all",
    # optional plot display
    plot_kwargs=None,
):
    """Plots multiple per-nucleotide datasets on a single axis.

    Parameters
    ----------
        samples : list of rnavigate Samples
            samples used to retrieve data
        profile : data keyword string or data object
            Profile from which values will be plotted
        sequence : data keyword str, data obj, or sequence str, defaults to `profile`
            All data are mapped to this sequence before plotting
            If a data keyword, data from the first sample will be used
        annotations : list of data keyword strings or data objects, defaults to []
            Annotations used to highlight regions or sites of interest
        domains : data keyword string or data object, defaults to None
            domains to label along x-axis
        labels : list of str, defaults to sample.sample for each sample
            list containing Labels to be used in plot legends
        nt_ticks : tuple of two integers, defaults to (20, 5)
            first integer is the gap between major tick marks
            second integer is the gap between minor tick marks
        columns : string or list of strings, defaults to profile.metric
            columns names of values from profile to plot
        errors : string or list of strings, defaults to None (no error bars)
            column names of error values for plotting error bars
        annotations_mode : "track" or "bars", defaults to "track"
            "track" will highlight annotations along the x-axis
            "bars" will use a vertical transparent bar over the plot
        seqbar : bool, defaults to ``True``
            whether to display the sequence along the x-axis
        region : list of 2 integers, defaults to [1, length of sequence]
            start and end positions to plot. 1-indexed, inclusive.
        plot_kwargs : dictionary, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.Skyline
            the skyline plot object
    """
    sequence = get_sequence(sequence, samples[0], profile)
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        alignment=sequence.null_alignment,
        profile=profile,
        annotations=annotations,
        domains=domains,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.Skyline")
    plot = plots.Skyline(
        num_samples=len(samples),
        nt_length=sequence.length,
        region=region,
        **plot_kwargs,
    )
    for data_dict in parsed_args.data_dicts:
        plot.plot_data(
            **data_dict,
            columns=columns,
            errors=errors,
            nt_ticks=nt_ticks,
            annotations_mode=annotations_mode,
            seqbar=seqbar,
        )
    plot.set_figure_size()
    plot.plot_colorbars()
    return plot


def plot_profile(
    # required
    samples,
    profile,
    # optional data inputs
    sequence=None,
    annotations=None,
    domains=None,
    # optional data display
    labels=None,
    nt_ticks=(20, 5),
    column=None,
    plot_error=True,
    annotations_mode="track",
    seqbar=True,
    region="all",
    # optional plot display
    colorbars=True,
    plot_kwargs=None,
):
    """Aligns reactivity profiles by sequence and plots them on seperate axes.

    Parameters
    ----------
        samples : list of rnavigate Samples
            samples used to retrieve data
        profile : data keyword string or data object
            Profile from which values will be plotted
        sequence : data keyword str, data obj, or sequence str, defaults to `profile`
            All data are mapped to this sequence before plotting
            If a data keyword, data from the first sample will be used
        annotations : list of data keyword strings or data objects, defaults to []
            Annotations used to highlight regions or sites of interest
        domains : data keyword string or data object, defaults to None
            domains to label along x-axis
        labels : list of strings, defaults to sample.sample for each sample
            list containing Labels to be used in plot legends
        nt_ticks : tuple of two integers, defaults to (20, 5)
            first integer is the gap between major tick marks
            second integer is the gap between minor tick marks
        column : string, defaults to profile.metric
            column name of values from profile to plot
        plot_error : bool, defaults to True
            Whether to plot error bars, values are determined by profile.metric
        annotations_mode : "track" or "bars", defaults to "track"
            "track" will highlight annotations along the x-axis
            "bars" will use a vertical transparent bar over the plot
        seqbar : bool, defaults to ``True``
            whether to display the sequence along the x-axis
        region : list of 2 integers, defaults to [1, length of sequence]
            start and end positions to plot. 1-indexed, inclusive.
        colorbars : bool, defaults to True
            Whether to plot color scales for per-nucleotide data
        plot_kwargs : dictionary, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.Profile
            the Profile plot object
    """
    sequence = get_sequence(sequence, samples[0], profile)
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        alignment=sequence.null_alignment,
        annotations=annotations,
        domains=domains,
        profile=profile,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.Profile")
    plot = plots.Profile(
        num_samples=len(samples),
        nt_length=sequence.length,
        region=region,
        **plot_kwargs,
    )
    for data_dict in parsed_args.data_dicts:
        plot.plot_data(
            **data_dict,
            column=column,
            plot_error=plot_error,
            seqbar=seqbar,
            annotations_mode=annotations_mode,
            nt_ticks=nt_ticks,
        )
    plot.set_figure_size()
    if colorbars:
        plot.plot_colorbars()
    return plot


def plot_alignment(
    # required
    data1,
    data2,
    # optional display
    labels=None,
    plot_kwargs=None,
):
    """Plots the sequence alignment used to compare two sequences

    Parameters
    ----------
        data1 : tuple (rnavigate Sample, data keyword string)
            a sample and data keyword to retrieve a sequence
        data2 : tuple (rnavigate Sample, data keyword string)
            another sample and data keyword to retrieve a second sequence
        labels : list of 2 strings, defaults to "sample.sample: data keyword" for each
            Labels used for each sample
        plot_kwargs : dict, defaults to {}
            passed to matplotlib.pyplot.subplots()

    Returns
    -------
        rnavigate.plots.Alignment
            the Alignment plot object
    """
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.Alignment")
    if labels is None:
        labels = [f"{s.sample}: {seq}" for s, seq in [data1, data2]]
    plot = plots.Alignment(num_samples=1, **plot_kwargs)
    alignment = data.SequenceAlignment(data1[0].data[data1[1]], data2[0].data[data2[1]])
    plot.plot_data(alignment=alignment, label=labels)
    plot.set_figure_size()
    return plot


def plot_arcs(
    # required
    samples,
    sequence,
    # optional data inputs
    structure=None,
    structure2=None,
    interactions=None,
    interactions2=None,
    profile=None,
    annotations=None,
    domains=None,
    # optional data display
    labels=None,
    nt_ticks=(20, 5),
    profile_scale_factor=1,
    plot_error=False,
    annotation_mode="track",
    panels=None,
    seqbar=True,
    region="all",
    # optional plot display
    colorbars=True,
    title=True,
    plot_kwargs=None,
):
    """Plots interactions and/or base-pairs as arcs.

    Parameters
    ----------
        samples : list of rnavigate Samples
            samples used to retrieve data
        sequence : data keyword string, data object, or sequence string
            All data are mapped to this sequence before plotting
            If a data keyword string, data from the first sample will be used
        structure : data keyword string or data object, defaults to None
            secondary structure to plot as arcs
        structure2 : data keyword string or data object, defaults to None
            another secondary structure to compare with the first structure
            arcs will be colored depending on which structure they are in
            Defaults to None
        interactions : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to plot as arcs, no filtering performed
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
            format 3 (list of format 2 dictionaries)
                This format allows multiple filtering schemes to be applied,
                each will be plotted on a seperate axis
        interactions2 : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to plot as arcs, no filtering performed
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
        profile : data or data keyword, defaults to None
            Profile from which values will be plotted
        annotations : list of data keyword strings or data objects, defaults to []
            Annotations used to highlight regions or sites of interest
        domains : data keyword string or data object, defaults to None
            domains to label along x-axis
        labels : list of strings, defaults to sample.sample for each sample
            list containing Labels to be used in plot legends
        nt_ticks : tuple of two integers, defaults to (20, 5)
            first integer is the gap between major tick marks
            second integer is the gap between minor tick marks
        profile_scale_factor : number, defaults to 1
            small profile values will be hard to see
            large profile values will overwhelm the plot
            e.g. use 1/10 to scale values down 10-fold, use 10 to scale up
        plot_error : bool, defaults to False
            Whether to plot error bars, values are determined by profile.metric
        annotation_mode : "track" or "bars", defaults to "track"
            "track" will highlight annotations along the x-axis
            "bars" will use a vertical transparent bar over the plot
        panels : dict, optional
            a dictionary of whether plot elements are displayed on the "top"
            (above x-axis) or "bottom" (below x-axis)
            Only the values you wish to change from the default are needed
            defaults to {"interactions": "bottom",
                         "interactions2": "bottom",
                         "structure": "top",
                         "profile": "top"}
        seqbar : bool, defaults to ``True``
            whether to display the sequence along the x-axis
        region : list of 2 integers, defaults to [1, length of sequence]
            start and end positions to plot. 1-indexed, inclusive.
        colorbars : bool, defaults to True
            Whether to plot colorbars for all plot elements
        title : bool, defaults to True
            Whether to display titles for each axis
        plot_kwargs : dict, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.AP
            the ArcPlot object
    """
    sequence = get_sequence(sequence, samples[0], structure)
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        alignment=sequence.null_alignment,
        annotations=annotations,
        domains=domains,
        structure=structure,
        structure2=structure2,
        profile=profile,
        interactions2=interactions2,
        interactions=interactions,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.AP")
    parsed_args.update_rows_cols(plot_kwargs)
    # initialize plot
    plot = plots.AP(
        num_samples=parsed_args.num_samples,
        nt_length=sequence.length,
        region=region,
        **plot_kwargs,
    )
    # loop through samples and interactions, adding each as a new axis
    for data_dict in parsed_args.data_dicts:
        plot.plot_data(
            **data_dict,
            sequence=sequence,
            title=title,
            panels=panels,
            plot_error=plot_error,
            annotation_mode=annotation_mode,
            seqbar=seqbar,
            profile_scale_factor=profile_scale_factor,
            nt_ticks=nt_ticks,
        )
    plot.set_figure_size()
    if colorbars:
        plot.plot_colorbars()
    return plot


def plot_arcs_compare(
    # required
    samples,
    sequence,
    # optional data inputs
    structure=None,
    structure2=None,
    interactions=None,
    interactions2=None,
    profile=None,
    # optional data display
    labels=None,
    profile_scale_factor=1,
    plot_error=False,
    region="all",
    # optional plot display
    colorbars=True,
    plot_kwargs=None,
):
    """Generates a single arc plot displaying combinations of secondary
    structures, per-nucleotide data, inter-nucleotide data, and sequence
    annotations. The first sample will be on top, the second on the bottom.
    Center shows how these sequences are being aligned. This view does not

    Parameters
    ----------
        samples : list of 2 rnavigate Samples
            samples used to retrieve data
            This plotting function can only compare two samples at a time
        sequence : data keyword string, data object, or sequence string
            All data are mapped to this sequence taken from their respective
            sample before plotting
        structure : data keyword string or data object, defaults to None
            secondary structure to plot as arcs
        structure2 : data keyword string or data object, defaults to None
            another secondary structure to compare with the first structure
            arcs will be colored depending on which structure they are in
        interactions : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to plot as arcs, no filtering performed
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
            format 3 (list of format 2 dictionaries)
                This format allows multiple filtering schemes to be applied,
                each will be plotted on a seperate axis
        interactions2 : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to plot as arcs, no filtering performed
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
        profile : data keyword string or data object, defaults to None
            Profile from which values will be plotted
        labels : list of strings, defaults to sample.sample for each sample
            list containing Labels to be used in plot legends
        profile_scale_factor : number, defaults to 1
            small profile values will be hard to see
            large profile values will overwhelm the plot
            e.g. use 1/10 to scale values down 10-fold, use 10 to scale up
        plot_error : bool, defaults to ``False``
            Whether to plot error bars, values are determined by profile.metric
        region : list of 2 integers, defaults to [1, length of sequence]
            start and end positions to plot. 1-indexed, inclusive.
        colorbars : bool, defaults to ``True``
            Whether to plot color scales for all plot elements
        plot_kwargs : dict, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.AP plot
            object containing matplotlib figure and axes with additional plotting
            and file saving methods
    """
    if len(samples) != 2:
        raise ValueError("Only 2 samples can be compared.")
    seq1 = get_sequence(sequence, samples[0], structure)
    seq2 = get_sequence(sequence, samples[1], structure)
    alignment = data.SequenceAlignment(seq1, seq2, full=True)
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        structure=structure,
        structure2=structure2,
        profile=profile,
        interactions2=interactions2,
        interactions=interactions,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.AP")
    # initialize plot
    plot = plots.AP(
        num_samples=1,
        nt_length=len(alignment.target_sequence),
        region=region,
        **plot_kwargs,
    )
    # loop through samples and filters, adding each as a new axis
    labels = []
    for data_dict, seq, panel in zip(parsed_args.data_dicts, [1, 2], ["top", "bottom"]):
        if seq == 1:
            seq, other_seq = seq1, seq2
        else:
            seq, other_seq = seq2, seq1
        alignment = data.SequenceAlignment(seq, other_seq, full=True)
        panels = {
            "structure": panel,
            "interactions": panel,
            "interactions2": panel,
            "profile": panel,
        }
        labels.append(data_dict.pop("label"))
        data_dict = fit_data(data_dict, alignment)
        plot.plot_data(
            ax=0,
            sequence=seq,
            track_height=6,
            label="",
            seqbar=False,
            panels=panels,
            **data_dict,
            plot_error=plot_error,
            profile_scale_factor=profile_scale_factor,
        )
    plots.plot_sequence_alignment(
        ax=plot.axes[0, 0],
        alignment=alignment.get_inverse_alignment(),
        labels=labels[::-1],
        top=6,
        bottom=0,
    )
    plot.set_figure_size()
    if colorbars:
        plot.plot_colorbars()
    return plot


def plot_ss(
    # required
    samples,
    structure,
    # optional data inputs
    profile=None,
    annotations=None,
    interactions=None,
    interactions2=None,
    # optional data display
    labels=None,
    colors=None,
    nt_ticks=None,
    bp_style="dotted",
    # optional plot display
    colorbars=True,
    plot_kwargs=None,
):
    """Generates a multipanel secondary structure drawing with optional
    coloring by per-nucleotide data and display of inter-nucleotide data and/or
    sequence annotations. Each plot may display a unique sample and/or
    inter-nucleotide data filtering scheme.

    Parameters
    ----------
        samples : list of rnavigate Samples
            samples used to retrieve data
        structure : data keyword string or data object
            secondary structure to plot as arcs
            All data are mapped to this sequence before plotting
        profile : data keyword string or data object, defaults to None
            Profile used for coloring if "profile" used in colors dictionary
        annotations : list of data keyword strings or data objects, defaults to []
            Annotations used to highlight regions or sites of interest
        interactions : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to plot on secondary structure, no filtering
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
            format 3 (list of format 2 dictionaries)
                This format allows multiple filtering schemes to be applied,
                each will be plotted on a seperate axis
        interactions2 : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to plot on secondary structure, no filtering
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
        labels : list of strings, defaults to sample.sample for each sample
            list containing Labels to be used in plot legends
            Defaults to sample.sample for each sample
        colors : dictionary, optional
            a dictionary of element: value pairs that determines how colors
            will be applied to each plot element and if that element is plotted
            only the elements you wish to change need to be included
            value options and what the colors represent:
                None: don"t plot this elelement
                "sequence": nucleotide identity
                "position": position in sequence
                "annotations": sequence annotations
                "profile": per-nucleotide data from profile
                    profile argument must be provided
                "structure": base-pairing status
                matplotlib color: all positions plotted in this color
                array of colors: a different color for each position
                    must be the same length as structure
            "sequence" may also use "contrast" which automatically chooses
                white or black, which ever contrasts better with "nucleotide"
                color
            Defaults to {"sequence": None,
                         "nucleotides": "sequence",
                         "structure": "grey",
                         "basepairs": "grey"}
        nt_ticks : integer, defaults to None (no labels)
            gap between major tick marks
        bp_style : "dotted", "line", or "conventional", defaults to "dotted"
            "dotted" plots basepairs as a dotted line
            "line" plots basepairs as a solid line
            "conventional" plots basepairs using Leontis-Westhof conventions
                for canonical and wobble pairs ("G-A" plotted as solid dot)
        colorbars : bool, defaults to True
            Whether to plot color scales for all plot elements
        plot_kwargs : dict, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.SS plot
            object containing matplotlib figure and axes with additional plotting and
            file saving methods
    """
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        structure=structure,
        annotations=annotations,
        profile=profile,
        interactions=interactions,
        interactions2=interactions2,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.SS")
    parsed_args.update_rows_cols(plot_kwargs)
    # initialize plot using all structure drawings
    plot = plots.SS(num_samples=parsed_args.num_samples, **plot_kwargs)
    # loop through samples and filters, adding each as a new axis
    for data_dict in parsed_args.data_dicts:
        data_dict = fit_data(data_dict, data_dict["structure"].null_alignment)
        plot.plot_data(**data_dict, colors=colors, nt_ticks=nt_ticks, bp_style=bp_style)
    plot.set_figure_size()
    if colorbars:
        plot.plot_colorbars()
    return plot


def plot_mol(
    # required
    samples,
    structure,
    # optional data inputs
    profile=None,
    interactions=None,
    # optional data display
    labels=None,
    style="cartoon",
    hide_cylinders=False,
    colors="grey",
    atom="O2'",
    rotation=None,
    orientation=None,
    get_orientation=False,
    # optional viewer display
    title=True,
    colorbars=True,
    width=400,
    height=400,
    rows=None,
    cols=None,
    background_alpha=1,
    show=True,
):
    """Generates a multipanel interactive 3D molecular rendering of a PDB
    structure. Nucleotides may be colored by per-nucleotide data or custom
    color lists. Inter-nucleotide data may be displayed as cylinders connecting
    atoms or residues. Each plot may display a unique sample and/or filtering
    scheme.

    Parameters
    ----------
        samples : list of rnavigate Samples
            samples used to retrieve data
        structure : data keyword string or data object
            3D structure to view as interactive molecule
            All data are mapped to this sequence before plotting
        profile : data keyword string or data object, defaults to None
            Profile used to color nucleotides if colors="profile"
        interactions : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to plot on molecule, no filtering performed
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
            format 3 (list of format 2 dictionaries)
                This format allows multiple filtering schemes to be applied,
                each will be plotted on a seperate axis
        labels : list of strings, defaults to sample.sample for each sample
            list containing Labels to be used in plot titles
        style : "cartoon", "cross", "line", "sphere" or "stick", defaults to "cartoon"
            sets the py3Dmol style for drawing the molecule
        hide_cylinders : bool, defaults to False
            whether to hide nucleotide cylinders (only shows backbone ribbon)
        colors : string or list of colors, defaults to "grey"
            "sequence": color by nucleotide identity
            "position": color by position in sequence
            "annotations": color by sequence annotations from `annotations`
            "profile": color by per-nucleotide data from `profile`
            "structure": color by base-pairing status
            matplotlib color: all positions plotted in this color
            array of colors: a different color for each position
                must be the same length as structure
        atom : string or dictionary, defaults to "O2'"
            which atoms to draw interactions between
            for DMS reactive atoms (N1 for A and G, N3 for U and C) use "DMS"
            use a dictionary to specify a different atom for each nucleotide
                e.g. "DMS" == {"A": "N1", "G": "N1", "U": "N3", "C": "N3"}
        rotation : dictionary, defaults to {"x": 0, "y": 0, "z": 0}
            axis-degrees pairs for setting the starting orientation of the
            molecule, only the axes to be rotated are needed
            e.g. {"x": 180} flips the molecule on the x-axis
        orientation : list of 9 floats, defaults to None
            set the precise starting orientation
            see get_orientation for more details
        get_orientation : bool, defaults to False
            allows getting the orientation for use with orientation argument
            all other arguments will be ignored and a larger, single panel view
            window is displayed with no title
                1. adjust the molecule to the desired orientation
                2. click on the molecule to display the orientation vector
                3. copy this orientation vector (manually)
                4. provide this list of values to the orientation argument
        title : bool, defaults to True
            whether to display the title
        colorbars : bool, defaults to True
            Whether to plot color scales for all plot elements
        width : integer, defaults to 400
            width of view window in pixels
        height : integer, defaults to 400
            height of view window in pixels
        rows : integer, defaults to None (set automatically)
            the number of rows in the view window
        cols : integer, defaults to None (set automatically)
            the number of columns in the view window
        background_alpha : float, defaults to 1 (completely opaque)
            the opacity of the view window, must be between 0 and 1
        show : bool, defaults to True
            whether to display the viewer object

    Returns
    -------
        rnavigate.plots.Mol:
            object containing py3dmol viewer with additional plotting and file saving
            methods
    """
    sequence = get_sequence(structure, samples[0])
    if get_orientation:
        plot = plots.Mol(num_samples=1, pdb=sequence, width=800, height=800)
        plot.plot_data(title=False, get_orientation=True)
        plot.view.show()
        return plot
    parsed_args = PlottingArgumentParser(
        samples=samples,
        alignment=sequence.null_alignment,
        labels=labels,
        profile=profile,
        interactions=interactions,
    )
    # initialize plot using 1st 3D structure (applies to all samples)
    structure = samples[0].get_data(structure)
    plot = plots.Mol(
        num_samples=parsed_args.num_samples,
        pdb=structure,
        width=width,
        height=height,
        background_alpha=background_alpha,
        rotation=rotation,
        orientation=orientation,
        style=style,
        rows=rows,
        cols=cols,
    )
    for data_dict in parsed_args.data_dicts:
        plot.plot_data(**data_dict, colors=colors, atom=atom, title=title)
    # hide nucleotide cylinders in all viewers
    if hide_cylinders:
        plot.hide_cylinders()
    # show viewer grid
    if show:
        plot.view.show()
    if colorbars:
        plot.plot_colorbars()
    return plot


def plot_heatmap(
    # required
    samples,
    sequence,
    # optional data input
    structure=None,
    interactions=None,
    regions=None,
    # optional data display
    labels=None,
    levels=None,
    interpolation="nearest",
    atom="O2'",
    plot_type="heatmap",
    weights=None,
    # optional plot display
    rows=None,
    cols=None,
    plot_kwargs=None,
):
    """Generates a multipanel plot displaying a heatmap of inter-nucleotide
    data (nucleotide resolution of 2D KDE) and/or contour map of pdb
    distances. Each plot may display a unique sample and/or filtering scheme.

    Parameters
    ----------
        samples : list of rnavigate Samples
            samples used to retrieve data
        sequence : data keyword string, data object, or sequence string
            All data are mapped to this sequence before plotting
        structure : data keyword string or data object, defaults to None
            secondary structure or 3D structure used to plot contour lines
            contour lines are drawn according to levels argument
        interactions : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to plot as a heatmap, no filtering performed
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
            format 3 (list of format 2 dictionaries)
                This format allows multiple filtering schemes to be applied,
                each will be plotted on a seperate axis
        regions : list of lists of 4 integers, defaults to None (no boxes)
            each inner list defines two regions of the RNA that are interacting
            a box will be drawn around this interaction on the heatmap
            e.g. [[10, 20, 50, 60], [35, 45, 70, 80]] draws 2 boxes
                the first box will connect nucleotides 10-20 and 50-60
                the second box will connect nucleotides 35-45 and 70-80
        labels : list of strings, defaults to sample.sample for each sample
            Labels to be used as titles, must be same length as samples list
        levels : list of floats, defaults to [5] contact distance or [20] 3D distance
            contours are drawn separating nucleotides above and below these
            distances
            if structure argument is a secondary structure
                distance refers to contact distance
            if structure argument is a 3D structure
                distance refers to spatial distance in angstroms
        interpolation : string, defaults to "nearest"
            one of matplotlib's interpolations for heatmap (used with imshow)
            "nearest" works well for shorter RNAs (under 300 nt)
            "none" works well for longer RNAs (over 1200 nt)
        atom : string or dictionary, defaults to "O2'"
            from which atoms to calculate distances
            for DMS reactive atoms (N1 for A and G, N3 for U and C) use "DMS"
            use a dictionary to specify a different atom for each nucleotide
                e.g. "DMS" == {"A": "N1", "G": "N1", "U": "N3", "C": "N3"}
        plot_type : "heatmap" or "kde", defaults to "heatmap"
            how to plot interactions data
            "heatmap" will plot raw data, each interaction is a pixel in a grid
            "kde" will calculate a kernel density estimate and plot 5 levels
        weights : string, defaults to None (no weights)
            weights to be used in kernel density estimation
            must be a column of interactions data
        rows : integer, defaults to None (determined automatically)
            number of rows of plots
        cols : integer, defaults to None (determined automatically)
            number of columns of plots
        plot_kwargs : dictionary, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.Heatmap
            object containing matplotlib figure and axes with additional plotting and
            file saving methods
    """
    sequence = get_sequence(sequence, samples[0], structure)
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        alignment=sequence.null_alignment,
        interactions=interactions,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.Heatmap")
    # initialize plot using 1st 3D pdb (applies to all samples)
    structure = samples[0].data[structure]
    plot = plots.Heatmap(
        parsed_args.num_samples, structure, rows=rows, cols=cols, **plot_kwargs
    )
    # loop through samples and interactions, adding each as a new axis
    for data_dict in parsed_args.data_dicts:
        plot.plot_data(
            **data_dict,
            regions=regions,
            levels=levels,
            weights=weights,
            interpolation=interpolation,
            atom=atom,
            plot_type=plot_type,
        )
    plot.set_figure_size()
    return plot


def plot_circle(
    # required
    samples,
    sequence,
    # optional data inputs
    structure=None,
    structure2=None,
    interactions=None,
    interactions2=None,
    annotations=None,
    profile=None,
    # optional data display
    colors=None,
    nt_ticks=(20, 5),
    gap=30,
    labels=None,
    # optional plot display
    colorbars=True,
    plot_kwargs=None,
):
    """Creates a figure containing a circle plot for each sample given.

    Data that can be plotted on circle plots includes annotations (highlights
    regions around the edge.Generates a multipanel secondary structure drawing
    with optional coloring by per-nucleotide data and display of inter-
    nucleotide data and/or sequence annotations. Each plot may display a unique
    sample and/or inter-nucleotide data filtering scheme.

    Parameters
    ----------
        samples : list of rnavigate Samples
            samples used to retrieve data
        sequence : data or data keyword
            All data are mapped to this sequence before plotting
        structure : data keyword string or data object, defaults to None
            Structure used to plot base-pairs on circle plot
        structure2 : data keyword str, data obj or list of either, defaults to None
            Structures to compare with Structure. Each base-pair is colored by
            which structure contains it or how many structures contain it.
        interactions : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to plot on cirle plot, no filtering
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
            format 3 (list of format 2 dictionaries)
                This format allows multiple filtering schemes to be applied,
                each will be plotted on a seperate axis
        interactions2 : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to plot on circle plot, no filtering
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
        annotations : list of data keyword strings or data objects, defaults to []
            Annotations used to highlight regions or sites of interest
        profile : data keyword string or data object, defaults to None
            Profile used for coloring if "profile" used in colors dictionary
        labels : list of strings, defaults to sample.sample for each sample
            list containing Labels to be used in plot legends
        colors : dictionary, optional
            a dictionary of element: value pairs that determines how colors
            will be applied to each plot element and if that element is plotted
            only the elements you wish to change need to be included
            Defaults to {"sequence": None,  # sequence not shown
                         "nucleotides": "sequence",
                         "structure": "grey"}
            value options and what the colors represent:
                None: don't plot this elelement
                "sequence": nucleotide identity
                "position": position in sequence
                "annotations": sequence annotations
                "profile": per-nucleotide data from profile
                    profile argument must be provided
                "structure": base-pairing status
                matplotlib color: all positions plotted in this color
                array of colors: a different color for each position
                    must be the same length as structure
            "sequence" may also use "contrast" which automatically chooses
                white or black, which ever contrasts better with "nucleotide"
                color
        nt_ticks : tuple of two integers, defaults to (20, 5)
            first integer is the gap between major tick marks
            second integer is the gap between minor tick marks
        gap : integer, defaults to 30
            Width of gap between 5' and 3' end in degrees
        colorbars : bool, defaults to True
            Whether to plot color scales for all plot elements
        plot_kwargs : dict, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.Circle
            object containing matplotlib figure and axes with additional plotting and
            file saving methods
    """

    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        sequence=sequence,
        structure=structure,
        structure2=structure2,
        interactions2=interactions2,
        profile=profile,
        annotations=annotations,
        interactions=interactions,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.Circle")
    parsed_args.update_rows_cols(plot_kwargs)
    plot = plots.Circle(num_samples=parsed_args.num_samples, **plot_kwargs)
    # loop through samples and interactions, adding each as a new axis
    for data_dict in parsed_args.data_dicts:
        data_dict = fit_data(data_dict, data_dict["sequence"].null_alignment)
        plot.plot_data(**data_dict, colors=colors, nt_ticks=nt_ticks, gap=gap)
    plot.set_figure_size()
    if colorbars:
        plot.plot_colorbars()
    return plot


def plot_linreg(
    # required
    samples,
    profile,
    # optional data inputs
    sequence=None,
    structure=None,
    annotations=None,
    # optional data display
    labels=None,
    kde=False,
    scale="linear",
    regression="pearson",
    colors="sequence",
    column=None,
    region="all",
    # optional plot display
    colorbars=True,
    plot_kwargs=None,
):
    """Performs linear regression analysis and generates scatter plots of all
    sample-to-sample profile vs. profile comparisons. Colors nucleotides by
    identity or base-pairing status.

    Parameters
    ----------
        samples : list of rnavigate Samples
            samples used to retrieve data
        profile : data keyword string or data object
            per-nucleotide data to perform linear regression
            all data are mapped to the sequence of the profile data from the
            first sample before plotting, unless sequence is supplied
        sequence : data keyword str, data obj, or sequence str, defaults to None
            a sequence from which to align all profiles
            if a data keyword, uses data from the first sample
        structure : data keyword string or data object, defaults to None
            Structure used for coloring if colors argument is "structure"
        annotations : list of data keyword strings or data objects, defaults to []
            Annotations used for coloring if colors argument is "annotations"
        labels : list of strings, defaults to sample.sample for each sample
            list containing Labels to be used in plot legends
        kde : bool, defaults to False
            whether to plot kde (density) instead of a scatter plot
        scale : "linear" or "log", defaults to "linear"
            "linear" performs regression on raw values, displays linear units
            "log" performs regression on log10(values), displays log10 units
        regression : "pearson" or "spearman", defaults to "pearson"
            "pearson" calculates Pearson R-squared (standard)
            "spearman" calculates Spearman R-squared (rank-order)
        colors : string or list of colors, defaults to "sequence"
            value options and what the colors represent:
                "sequence": nucleotide identity
                "position": position in sequence
                "annotations": sequence annotations
                "profile": per-nucleotide data from profile
                    profile argument must be provided
                "structure": base-pairing status
                matplotlib color: all positions plotted in this color
                array of colors: a different color for each position
                    must be the same length as structure
        column : string, defaults to profile.metric
            column name of values from profile to use in regression
        region : list of 2 integers, defaults to [1, length of sequence]
            start and end nucleotide positions to include. 1-indexed, inclusive
        colorbars : bool, defaults to ``True``
            Whether to plot colorbars for scatter plot colors
        plot_kwargs : dict, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.LinReg
            object containing matplotlib figure and axes with additional plotting and
            file saving methods
    """
    sequence = get_sequence(sequence, samples[0], profile)
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        alignment=sequence.null_alignment,
        structure=structure,
        profile=profile,
        annotations=annotations,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.LinReg")
    plot = plots.LinReg(
        num_samples=parsed_args.num_samples,
        scale=scale,
        regression=regression,
        kde=kde,
        region=region,
        **plot_kwargs,
    )
    for data_dict in parsed_args.data_dicts:
        plot.plot_data(**data_dict, colors=colors, column=column)
    plot.set_figure_size()
    if colorbars:
        plot.plot_colorbars()
    return plot


def plot_roc(
    # required
    samples,
    structure,
    profile,
    # optional data display
    labels=None,
    nts="AUCG",
    # optional plot display
    plot_kwargs=None,
):
    """Performs receiver operator characteristic analysis (ROC), calculates
    area under ROC curve (AUC), and generates ROC plots to assess how well
    per-nucleotide data predicts base-paired status. Does this for all
    positions as well as positions categorized by nucleotide
    5 plots: All, A, U, C, G

    Parameters
    ----------
        samples : list of rnavigate Samples
            samples used to retrieve data
        structure : data keyword string or data object
            secondary structure to use as classifier (paired or unpaired)
            profile data for each sample is first aligned to this structure
        profile : data keyword string or data object
            per-nucleotide data to perform ROC analysis
        labels : list of strings, defaults to sample.sample for each sample
            list containing Labels to be used in plot legends
        nts : string, defaults to "AUCG"
            which nucleotides to plot nucleotide-type ROC plots
        plot_kwargs : dict, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.ROC
            object containing matplotlib figure and axes with additional plotting and
            file saving methods
    """
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        structure=structure,
        profile=profile,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.ROC")
    plot = plots.ROC(parsed_args.num_samples, **plot_kwargs)
    for data_dict in parsed_args.data_dicts:
        data_dict = fit_data(data_dict, data_dict["structure"].null_alignment)
        plot.plot_data(**data_dict, nts=nts)
    plot.set_figure_size()
    return plot


def plot_disthist(
    # required
    samples,
    structure,
    interactions,
    # optional data inputs
    bg_interactions=None,
    # optional data display
    labels=None,
    same_axis=False,
    atom="O2'",
    # optional plot display
    rows=None,
    cols=None,
    plot_kwargs=None,
):
    """Calculates 3D distance of nucleotides in inter-nucleotide data and plots
    the distribution of these distances. Compares this to a "background"
    distribution consisting of either all pairwise distances in structure, or
    those defined by bg_interactions and bg_interactions_filter

    Parameters
    ----------
        samples : list of rnavigate Samples
            Samples from which to retreive data
            There will be one panel for each sample unless same_axis is True
        structure : data keyword string or data object
            secondary structure or 3D structure to calculate inter-nucleotide
            contact distance or 3D distance, respectively
        interactions : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions used to calculate distance histogram, no filtering
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
            format 3 (list of format 2 dictionaries)
                This format allows multiple filtering schemes to be applied,
                each will be plotted on a seperate axis
        bg_interactions : one of the formats below, defaults to None
            format 1 (data or data keyword)
                Interactions to calculate background distance histogram, no
                filtering is performed
            format 2 (dictionary)
                e.g. {"interactions": format 1}
                additional filtering options can be added to the dictionary
            if not provided, background distance histogram is calculated from
            all pairwise distances in structure
        labels : list of strings, defaults to sample.sample for each sample
            Labels to be used as titles, must be same length as samples list
            Defaults to sample.sample for each sample
        atom : string or dictionary, defaults to "O2'"
            from which atoms to calculate distances
            for DMS reactive atoms (N1 for A and G, N3 for U and C) use "DMS"
            use a dictionary to specify a different atom for each nucleotide
                e.g. "DMS" == {"A": "N1", "G": "N1", "U": "N3", "C": "N3"}
        rows : integer, defaults to None (determined automatically)
            number of rows of plots
        cols : integer, defaults to None (determined automatically)
            number of columns of plots
        plot_kwargs : dictionary, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.DistHist
            object containing matplotlib figure and axes with additional plotting and
            file saving methods
    """
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        structure=structure,
        bg_interactions=bg_interactions,
        interactions=interactions,
    )
    plot_kwargs = _parse_plot_kwargs(plot_kwargs, "rnavigate.plots.DistHist")
    # initialize plot
    if same_axis:
        plot = plots.DistHist(parsed_args.num_samples, rows=1, cols=1, **plot_kwargs)
        ax = plot.axes[0, 0]
    else:
        plot = plots.DistHist(
            parsed_args.num_samples, rows=rows, cols=cols, **plot_kwargs
        )
        ax = None
    # loop through samples and interactions, adding each as a new axis
    for data_dict in parsed_args.data_dicts:
        plot.plot_data(**data_dict, ax=ax, atom=atom)
    plot.set_figure_size()
    return plot


def plot_ntdist(
    # required
    samples,
    profile,
    # optional data display
    labels=None,
    column=None,
    # optional plot display
    plot_kwargs=None,
):
    """Plots the distributions of values at A, U, C, and G.

    Calculates the kernel density estimate (KDE) for each nucleobase and plots
    them on one axis per sample.

    Parameters
    ----------
        samples : list of rnavigate Samples
            samples used to retrieve data
        profile : data keyword string or data object
            per-nucleotide data to plot per-nt-identity distributions
        labels : list of strings, defaults to sample.sample for each sample
            list containing Labels to be used in plot legends
        column : string, defaults to profile.metric
            which column of data to use for KDE
        plot_kwargs : dict, defaults to {}
            Keyword-arguments passed to matplotlib.pyplot.subplots

    Returns
    -------
        rnavigate.plots.NucleotideDistribution
            object containing matplotlib figure and axes with additional
            plotting and file saving methods
    """
    parsed_args = PlottingArgumentParser(
        samples=samples,
        labels=labels,
        profile=profile,
    )
    plot_kwargs = _parse_plot_kwargs(
        plot_kwargs, "rnavigate.plots.NucleotideDistribution"
    )
    plot = plots.NucleotideDistribution(parsed_args.num_samples, **plot_kwargs)
    for data_dict in parsed_args.data_dicts:
        plot.plot_data(**data_dict, column=column)
    plot.set_figure_size()
    return plot



"""SequenceChecker analysis used to inspect sequence differences.

Given a list of samples, we can inspect which data keywords belong to the
samples, which sequences match up perfectly, and inspect the differences
between sequences.
"""
import pandas as pd
import numpy as np
from rnavigate import data


class SequenceChecker:
    """Check the sequences stored in a list of samples.

    Attributes:
        samples: the list of samples
        sequences: a list of all unique sequence strings stored in the list of
            samples. These are converted to an all uppercase RNA alphabet.
        keywords: a list of unique data keywords stored in the list of samples.
        which_sequences: a dataframe of samples and keywords and which
            sequences each contains.
    """

    def __init__(self, samples):
        """Creates an instance of SequenceChecker given a list of samples

        Arguments:
            samples (list of rnav.Sample)
                samples for which to compare data keywords and sequences.
        """
        self.samples = samples
        self.keywords = self.get_keywords()
        self.sequences = self.get_sequences()
        self.which_sequences = self.get_which_sequences()

    def reset(self):
        """Reset keywords and sequences from sample list in case of changes."""
        self.keywords = self.get_keywords()
        self.sequences = self.get_sequences()
        self.which_sequences = self.get_which_sequences()

    def get_keywords(self):
        """A list of all unique data keywords across samples."""
        return list(set([dkw for s in self.samples for dkw in s.data]))

    def get_sequences(self):
        """A list of all unique sequences (uppercase RNA) across samples."""
        sequences = []
        for sample in self.samples:
            for dkw in self.keywords:
                if dkw not in sample.data:
                    continue
                seq = sample.get_data(dkw).sequence
                seq = seq.upper().replace("T", "U")
                if seq not in sequences:
                    sequences.append(seq)
        return sequences

    def get_which_sequences(self):
        """A DataFrame of sequence IDs (integers) for each data keyword."""
        df = {key: [] for key in ["Sample"] + self.keywords}
        for sample in self.samples:
            df["Sample"].append(sample.sample)
            for dkw in self.keywords:
                if dkw not in sample.data:
                    df[dkw].append(np.nan)
                else:
                    seq = sample.get_data(dkw).sequence
                    seq = seq.upper().replace("T", "U")
                    df[dkw].append(self.sequences.index(seq))
        return pd.DataFrame(df)

    def print_which_sequences(self):
        """Print sequence ID (integer) for each data keyword and sample."""
        print("Sequence IDs")
        for _, row in self.which_sequences.iterrows():
            for column in self.which_sequences.columns:
                if column == "Sample":
                    print(f"    {row[column]}")
                else:
                    if np.isnan(row[column]):
                        continue
                    print(f"        {column:<10} {int(row[column])}")
        print()

    def print_alignments(self, print_format="long", which="all"):
        """Print alignments in the given format for sequence IDs provided.

        Optional arguments:
            print_format (string)
                What format to print the alignments in:
                "cigar" prints the cigar string
                "short" prints the numbers of mismatches and indels
                "long" prints the location and nucleotide identity of all
                    mismatches, insertions and deletions.
                Defaults to "long".
            which (pair of integers)
                two sequence IDs to compare.
                Defaults to every pairwise comparison.
        """
        kwargs = {"print_format": print_format}
        if which == "all":
            num = range(len(self.sequences))
            for s1 in num:
                for s2 in num:
                    if s1 < s2:
                        self.print_alignments(which=(s1, s2), **kwargs)
            return
        i, j = int(which[0]), int(which[1])
        a = self.sequences[i]
        b = self.sequences[j]
        alignment = data.SequenceAlignment(sequence1=a, sequence2=b, full=True)
        print(f"Sequence {i} -> {j} ({len(a)} nts -> {len(b)} nts)")
        alignment.print(print_format=print_format)

    def print_mulitple_sequence_alignment(self, base_sequence):
        """Print the multiple sequence alignment with nice formatting.

        Required arguments:
            base_sequence (string)
                a sequence string that represents the longest common sequence.
                Usually, this is the return value from:
                    rnav.data.set_multiple_sequence_alignment()
        """
        print("Multiple sequence alignment")
        alignments = [
            data.SequenceAlignment(base_sequence, seq).alignment2
            for seq in self.sequences
        ]
        pos = "".join(f"{n:<20}" for n in range(1, len(alignments[0]), 20))
        misses = "".join("X "[len(set(nts)) == 1] for nts in zip(*alignments))
        print("    ID    length   alignment")
        for i, (alignment, seq) in enumerate(zip(alignments, self.sequences)):
            print(f"    {i:<5} {len(seq):<8} {alignment}")
        print(f"    {'mismatches':<14} {misses}")
        print(f"    {'positions':<14} {pos}")
        print()

    def write_fasta(self, filename, which="all"):
        """Write all unique sequences to a fasta file.

        This is very useful for using external multiple sequence aligners such
        as ClustalOmega.
            1) go to https://www.ebi.ac.uk/Tools/msa/clustalo/
            2) upload new fasta file
            3) under STEP 2 output format, select Pearson/FASTA
            4) click 'Submit'
            5) wait for your alignment to finish
            6) download the alignment fasta file
            7) use rnav.data.set_multiple_sequence_alignment()

        Required arguments:
            filename (string)
                path to a new file to which fasta entries are written

        Optional arguments:
            which (list of integers)
                Sequence IDs to write to file.
        """
        if which == "all":
            which = range(len(self.sequences))
        with open(filename, "w") as fasta:
            for i in which:
                fasta.write(f">Sequence_{i}\n")
                fasta.write(f"{self.sequences[i]}\n")



from matplotlib.collections import PatchCollection
from matplotlib.patches import Path, PathPatch
import numpy as np
from rnavigate import styles


def plot_interactions_circle(ax, seq_circle, interactions):
    """Plot interactions on a circle plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot on.
    seq_circle : rnavigate.data.SequenceCircle
        The sequence circle object containing nucleotide positions.
    interactions : rnavigate.data.Interactions
        The interactions to be plotted as arcs between nucleotides.
    """
    radius = seq_circle.radius
    theta = seq_circle.data["Theta"]
    ij_colors = interactions.get_ij_colors()
    patches = []
    for i, j, color in zip(*ij_colors):
        if j < i:
            i, j = j, i
        theta_i = theta[i - 1]
        theta_j = theta[j - 1]
        # scaling the center point towards zero depending on angle(i,j)
        theta_center = (theta_i + theta_j) / 2
        theta_diff = theta_j - theta_i
        if theta_diff > np.pi:
            theta_diff = 2 * np.pi - theta_diff
            theta_center -= np.pi
        r_center = radius * (1 - (theta_diff / 2 / np.pi)) ** 4
        verts = [[theta_i, radius], [theta_center, r_center], [theta_j, radius]]
        codes = [Path.MOVETO, Path.CURVE3, Path.CURVE3]
        patches.append(PathPatch(Path(verts, codes), fc="none", ec=color))
    ax.add_collection(PatchCollection(patches, match_original=True))


def plot_annotation_circle(ax, seq_circle, annotation, offset=1):
    """Plot annotations on a circle plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot on.
    seq_circle : rnavigate.data.SequenceCircle
        The sequence circle object containing nucleotide positions.
    annotation : rnavigate.data.Annotation
        The annotation to be plotted.
    offset : float, optional
        The offset from the circle circumference to plot the annotation.
    """
    radius = seq_circle.radius + offset
    color = annotation.color
    if annotation.annotation_type in ["spans", "primers"]:
        for start, end in annotation:
            if start > end:
                start, end = end, start
            theta = seq_circle.data.loc[start - 1 : end, "Theta"]
            if start == end:
                ax.scatter(theta, radius, color=color, **styles.settings["ss"]["sites"])
            else:
                ax.plot(theta, radius, color=color, **styles.settings["ss"]["spans"])
    elif annotation.annotation_type in ["sites", "group"]:
        sites = annotation.data["Site"] - 1
        theta = seq_circle.data.loc[sites, "Theta"]
        if annotation.annotation_type == "group":
            start, end = min(sites), max(sites)
            thetas = seq_circle.data.loc[start : end - 1, "Thetas"]
            ax.plot(thetas, radius, color=color, **styles.settings["ss"]["spans"])
        ax.scatter(theta, radius, color=color, **styles.settings["ss"]["sites"])



from rnavigate.plots.plots import Plot
from rnavigate.plots.alignment import Alignment
from rnavigate.plots.profile import Profile
from rnavigate.plots.ss import SS

from rnavigate.plots.functions import (
    get_contrasting_colors,
    adjust_spines,
    clip_spines,
    get_nt_ticks,
    set_nt_ticks,
    box_xtick_labels,
    plot_interactions_arcs,
    plot_profile_bars,
    plot_profile_skyline,
    plot_sequence_alignment,
    plot_annotation_track,
    plot_domain_track,
    plot_sequence_track,
    plot_annotation_ss,
    plot_basepairs_ss,
    plot_interactions_ss,
    plot_nucleotides_ss,
    plot_positions_ss,
    plot_sequence_ss,
    plot_structure_ss,
    plot_annotation_circle,
    plot_interactions_circle,
)

__all__ = [
    "Alignment",
    "AP",
    "Circle",
    "Profile",
    "QC",
    "SS",
    "get_contrasting_colors",
    "adjust_spines",
    "clip_spines",
    "get_nt_ticks",
    "set_nt_ticks",
    "box_xtick_labels",
    "plot_interactions_arcs",
    "plot_profile_bars",
    "plot_profile_skyline",
    "plot_sequence_alignment",
    "plot_annotation_track",
    "plot_domain_track",
    "plot_sequence_track",
    "plot_annotation_ss",
    "plot_basepairs_ss",
    "plot_interactions_ss",
    "plot_nucleotides_ss",
    "plot_positions_ss",
    "plot_sequence_ss",
    "plot_structure_ss",
    "plot_annotation_circle",
    "plot_interactions_circle",
]



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Module for parsing fpocket outputs and analyzing pocket characteristics
# Seth Veenbaas
# Weeks Lab, UNC-CH
# 2022
#
# Version 1.3.0
#
# -----------------------------------------------------------------------------
import os
import math
from re import findall, sub
from prody import *
import numpy as np
import pandas as pd
import requests
from pymol import cmd
from rdkit import Chem
from rdkit.Chem import QED
import trimesh
from fpocketR import util
from prody.utilities import openFile


def analyze_pockets(
    pdb : str,
    pqr_out : str,
    pdb_out : str,
    analysis : str,
    name : str,
    info_txt : str,
    pockets_out : list[str],
    pdb_code : str,
    chain : str,
    state,
    ligandchain : str,
    ligand : str,
    m : float,
    M : float,
    i : int,
    D : float,
    A : int,
    p : float,
    qualityfilter : float,
    knownnt : list[int],
    ) -> tuple[pd.DataFrame, prody.AtomGroup]:

    # Parses pdb files and returns prody structure objects.
    ligand_rna_structure = parsePDB(pdb)
    out_rna_structure = parsePDB(pdb_out)
    pocket_structure = parsePQR(pqr_out)

    # Sets ligand chain to first pdb chain by default.
    if ligandchain is None:
        ligandchain = chain[0]

    # Create real_sphere.pdb ouput be combinding the pqr_out and pdb_out.
    get_real_sphere(pqr_out, pdb_out, analysis, name)

    # Creates a dataframe with pocket characteristics from fpocket info.txt
    pc_df = get_characteristics(
        info_txt,
        pdb_code,
        name,
        m,
        M,
        i,
        D,
        A,
        p,
        state,
    )

    rna_coords = out_rna_structure.copy()

    # If pockets are detected calculate features and add to pc_df.
    if out_rna_structure.select('resname STP'):

        # Get atomgroup for rna and add pocket characteristics.
        stp_coords = out_rna_structure.select('resname STP').copy()

        add_basic_characteristics(
            stp_coords,
            pockets_out,
            qualityfilter,
            pc_df,
            chain,
            analysis,
            name,
            knownnt,
        )
        
        # Get atomgroup for ligand and add ligand characteristics.
        if ligand in ('n', 'N', 'no', 'No', 'none', 'None'):
            ligand_coords = None
            ligand = None
        else:
            print(f'Ligand: {ligand}')
            ligand_coords, ligand = get_ligand_coords(
                ligand_rna_structure,
                ligand,
                ligandchain,
                analysis,
                name,
            )
            if ligand_coords:
                add_ligand_characteristics(
                    analysis,
                    stp_coords,
                    ligand_coords,
                    ligand,
                    pc_df,
                )

    return pc_df, rna_coords


# -----------------------------------------------------------------------------


def reformat_pqr(filename : str):
    """Reformats PQR files to add whitespace between all data fields.
    This is required for Prody 2.4 to properly parse PQR files.
    NOTE: reformated PQR files will not be read properly by PyMol 2.5.

    Args:
        filename (str): Path to PQR file.

    Returns:
        str: Path to reformated PQR file. <filename>_prody.pqr
    """
    pqr = openFile(filename, 'rt')
    lines = pqr.readlines()
    pqr.close()
    reformat=[]
    atom = False

    for line in lines:
        startswith = line[0:6].strip()

        if startswith == 'ATOM' or startswith == 'HETATM':
            atom = True
            fields = line.split()
            if fields[5].find('.') != -1:
            # coords too early as no chid
                fields.insert(4, '')
            if len(fields) != 11:
                try:
                    fields = fields[:6] + [line[30:38].strip(), line[38:46].strip(), line[46:54].strip()] + line[54:].split()
                except:
                    continue
            
            serial_str = fields[1]
            atomname= fields[2]
            resname = fields[3]
            chid = fields[5]
            alt = ' '
            coordinates_x = fields[6]
            coordinates_y = fields[7]
            coordinates_z = fields[8]
            charges = fields[9]
            radii = fields[10]

            new_line = (f'{startswith} {serial_str} {atomname} {resname} '
                        f'{chid} {alt} {coordinates_x} {coordinates_y} '
                        f'{coordinates_z} {charges} {radii}')
        else:
            new_line = line

        reformat.append(new_line)
    if atom is True:
        prody_pqr = f'{filename[:-4]}_prody.pqr'
        with open(prody_pqr, 'w') as f:
            for new_line in reformat:
                f.write(f"{new_line}\n")
        return prody_pqr
    else: 
        return None
    
    
def get_real_sphere(pqr_file, pdb_file, analysis, name):
    """Encodes fpocket pocket a-sphere radii into a single .pdb output file.
    a-sphere radii incoded into the B factor column of the output .pdb file.
    Args:
        pqr_file (string): Path to fpocket *out.pqr file
                           containing a-spheres radii.
        pdb_file (string): Path to fpocket *out.pdb file.
    """
    # Opens .pqr and .pdb inputs and .pdb output.
    with open(pqr_file, 'r') as pqr_file, open(pdb_file, 'r') as pdb_file, \
            open(f'{analysis}/{name}_out_real_sphere.pdb', 'a') as out:
        pqr_line = pqr_file.readline()
        pdb_lines = pdb_file.readlines()
        # Iterates through Header lines in pqr file
        while pqr_line.startswith("HEADER"):
            pqr_line = pqr_file.readline()
        # Iternates through each line of the pdb file.
        for pdb_line in pdb_lines:
            if pdb_line.startswith("HEADER") or pdb_line.startswith("ATOM"):
                line_str = pdb_line[0:80]
            elif pdb_line.startswith("HETATM") and 'APOL' not in pdb_line:
                line_str = pdb_line[0:80]
            elif pdb_line.startswith("HETATM") and \
                    pdb_line[24:55] == pqr_line[24:55]:
                line_str = pdb_line[0:60] + \
                    pqr_line[65:71] + pdb_line[66:80]
                # Advance to next line of pqr file.
                pqr_line = pqr_file.readline()
            else:
                print(f"ERROR: pqr_to_pdb. \n"
                      f"PDB input file: {pdb_file} \n"
                      f"PQR input file: {pqr_file}")
                break
            # Write line to out_real_sphere.pdb file
            out.write(line_str)
            out.write('\n')


def get_ligand_coords(
    ligand_rna_structure : prody.AtomGroup,
    ligand : str,
    ligandchain : str,
    analysis : str,
    name : str,
) -> tuple[prody.AtomGroup, str]:
    """Gets coordinates and residue name for RNA-binding ligand.

    Args:
        ligand_rna_structure (object): Prody structure of RNA-ligand complex.
        ligand (str): Ligand residue name (usually a 3-letter code).
        ligandchain (str): Chain identifier for desired ligand.
        analysis (str): Path directory contianing fpocket outputs for analysis.
        name (str): Name of input pdb file.

    Returns:
        object: ProDy atomgroup of all atoms in known RNA ligand.
        str: Ligand residue name (usually a 3-letter code).
    """
    if ligand and 2 <= len(ligand) <= 3:
        try:
            ligand_sel = ligand_rna_structure.select(
                f'chain {ligandchain} and resname {ligand}').copy()
            if ligand_sel is None:
                return (None, None)
            else:
                ligand_coords = ligand_sel.copy()
                return (ligand_coords, ligand)
        except AttributeError:
            print(f'Chain {ligandchain} of {name}.pdb does not contain a '
                  f'ligand named: {ligand}\n'
                  'Please provide a valid ligand chain (--ligandchain) '
                  'and ligand residue name (--ligand).')
            return (None, None)

    elif ligand_rna_structure.select(f'chain {ligandchain} and hetatm '
                                 'and not ion and not water') is None:
        return (None, None)

    else:
        ligand_sele = ligand_rna_structure.select(
            f'chain {ligandchain} and hetatm and not ion and not water').copy()
        hetatm_resn : list = np.unique(ligand_sele.getResnames()).tolist()
    
        if len(hetatm_resn) > 1:
            resnames_qeds: dict = {}
            for resname in hetatm_resn:
                try:
                    response = requests.get(
                        f'https://www.ebi.ac.uk/pdbe/static/files/pdbechem_v2/{resname}_ideal.sdf')
                    with open(f'{analysis}/{resname}_ideal.sdf', 'wb') as f:
                        f.write(response.content)
                    mol = Chem.MolFromMolFile(f'{analysis}/{resname}_ideal.sdf')
                    qed = QED.default(mol)
                    mw : float = Chem.rdMolDescriptors.CalcExactMolWt(mol)
                    pat = Chem.MolFromSmarts("[#6]")
                    num_of_carbon : int =len(mol.GetSubstructMatches(pat))
                    if mw < 100.0 or num_of_carbon <= 3:
                        qed = 0
                        continue
                        
                except:
                    print(f'Error: Not able to calculate QED score for {resname}.\n')
                    qed = 0
                
                resnames_qeds[resname] = qed
                
            # Get ligand with the highest qed score
            hetatm_resn = [max(resnames_qeds, key=resnames_qeds.get)]

        if len(hetatm_resn) == 1 \
                and 2 <= len(hetatm_resn[0]) <= 3:
            ligand_coords = ligand_sele.select(
                f'chain {ligandchain} and resname {hetatm_resn[0]}').copy()
            print(f'Using {hetatm_resn[0]} as ligand for analysis.')
            ligand = hetatm_resn[0]
            print(ligand_coords, ligand)
            return (ligand_coords, ligand)

        elif len(hetatm_resn) == 1 and 2 > len(hetatm_resn[0]) > 3:
            print('No ligand detected.\n')
            return (None, None)

        elif len(hetatm_resn) > 0:
            while True:
                input_resn = input(
                    f'Detected heteroatoms: {hetatm_resn}.\n\n'
                    'Input the target ligand ID (case-sensitive; "none" for no ligand): ')
                print('\n')

                if input_resn in ('n', 'N', 'no', 'No', 'none', 'None'):
                    return (None, None)
                elif input_resn in hetatm_resn:
                    ligand_coords = ligand_sele.select(
                        f'chain {ligandchain} and resname {input_resn}').copy()
                    print(f'Using {input_resn} as ligand for analysis.')
                    ligand = input_resn
                    break
                else:
                    print(f'NameError: {input_resn} is not a not a valid '
                          'ligand ID.\n'
                          'Input "None" to proceed without a ligand.\n')

        else:
            print('No ligand detected.\n')
            return (None, None)


def get_characteristics(
    info_txt : str,
    pdb_code :str,
    name : str,
    m : float,
    M : float,
    i : int,
    D : float,
    A : int,
    p : float,
    state : int,
) -> pd.DataFrame:
    """Creates a dataframe containing characteristics for
        all fpocket generates pockets.

    Args:
        info_txt (string): path to *info.txt file outputed by fpocket.
        pdb_code (string): 4 digit identifier for the PDB structure.
        m (float): Specifies the minimum radius for an a-sphere.
        M (float): Specifies the maximum radius for an a-sphere.
        i (int): Specifies the minimum number of a-spheres per pocket.
        D (float): Specifies the a-sphere clustering distance for pockets.
        A (int): # of electroneg. atoms to define a polar a-sphere.
        p (float): Max. ratio of apolar a-spheres in a pocket.
        state (int): Structural state to analyze.

    Returns:
        DataFrame: Displays the characteristics and properties most relvant
        to scoring each pocket.
    """

    columns = [
        'Parameters', 'Name', 'PDB', 'State', 'Type', 'Filter', 'Pocket', 'Score',
        'Drug_score', 'a-sphere', 'SASA', 'Volume', 'Hydrophobic_density',
        'Apolar_a-sphere_proportion', 'Hydrophobicity_score', 'Polarity_score',
        'PocketNT', 'Pocket_NPR1', 'Pocket_NPR2', 'Pocket_shape', 'Ligand_ID',
        'Pocket_overlap', 'Ligand_overlap', 'Center_criteria', 'QED_score',
        'Ligand_NPR1', 'Ligand_NPR2', 'Ligand_shape'
    ]

    pc_d = {col: [] for col in columns}

    with open(info_txt, 'r') as f:

        poc_count = 1

        for row in f:
            if 'Pocket' in row:
                pc_d['Parameters'].append(f'-m {m} -M {M} -i {i} -D {D} '
                                          f'-A {A} -p {p}')
                pc_d['Name'].append(name)
                pc_d['PDB'].append(pdb_code)
                pc_d['State'].append(state)
                pc_d['Pocket'].append(int(row.split(' ')[1].strip()))
                pocket = int(row.split(' ')[1].strip())
            if 'Score :' in row:
                if 'Druggability' not in row:
                    pc_d['Score'].append(float(row.split(':')[1].strip()))
            if 'Druggability Score :' in row:
                pc_d['Drug_score'].append(float(row.split(':')[1].strip()))
            if 'Number of Alpha Spheres :' in row:
                pc_d['a-sphere'].append(int(row.split(':')[1].strip()))
            if 'Total SASA :' in row:
                pc_d['SASA'].append(float(row.split(':')[1].strip()))
            if 'Volume :' in row:
                pc_d['Volume'].append(float(row.split(':')[1].strip()))
            if 'Mean local hydrophobic density' in row:
                pc_d['Hydrophobic_density'].append(
                    float(row.split(':')[1].strip()))
            if 'Apolar alpha sphere proportion' in row:
                pc_d['Apolar_a-sphere_proportion'].append(
                    float(row.split(':')[1].strip()))
            if 'Hydrophobicity score' in row:
                pc_d['Hydrophobicity_score'].append(
                    float(row.split(':')[1].strip()))
            if 'Polarity score:' in row:
                pc_d['Polarity_score'].append(float(row.split(':')[1].strip()))
            if 'Flexibility' in row and poc_count == pocket:
                poc_count += 1

    # Ensure all columns have the same length
    max_len = max(len(v) for v in pc_d.values())
    for col in columns:
        while len(pc_d[col]) < max_len:
            pc_d[col].append(None)

    pc_df = pd.DataFrame.from_dict(pc_d)
    pc_df['Filter'] = 'Fail'
    pc_df['Type'] = 'Novel'

    return pc_df


def add_basic_characteristics(
    stp_coords: prody.AtomGroup,
    pockets_out: list[str],
    qualityfilter: float,
    pc_df: pd.DataFrame,
    chain: str,
    analysis: str,
    name: str,
    knownnt: list[int],
) -> None:
    """Adds characteristics to the pocket characteristics DataFrame that do
        not require a ligand to calculate.
    PocketNT: nucleotides in contact with pocket,
    Geometric Center: X,Y,Z coordinates of pocket center,
    Filter: Pocket quality filter (Pass or Fail)
    
    Args:
        stp_coords (object): ProDy atom group of a-sphere coordinates.
        pockets_out list[str]: Valid paths to pockets/pocket*_atm.pdb files.
        qualityfilter (float): Minimum fpocket score filter for pockets.
        pc_df (DataFrame): Characteristics and properities for each pocket.
        chain (str): Chain identifier for desired RNA chain (default='A').
        analysis (str): path directory contianing fpocket outputs for analysis.
        name (str): Name of input pdb file.
    """
    pocketNT = []
    pocket_npr1 = []
    pocket_npr2 = []

    for i, _ in enumerate(stp_coords.iterResidues()):
        
        # Calculate pocketNT (nucleotides near each pocket).
        x = pockets_out[i]
        with open(x, 'r') as f:
            tmp0 = f.read()
            pc_dict = {'Pocket': [(int(x[1]))] for x in findall(
                r'(\w+\s\w+\s*)(\d)(:)(\n)', tmp0)}

            pc_dict.update({x[1].strip(): float(x[2]) for x in findall(
                r'(\w+\s\w+\s*-\s*)(.+):\s*([\d.-]+)(\n)', tmp0)})

        structure = parsePDB(x)

        if ',' in chain:
            chain = chain.replace(',', ' & ')
        selection = structure.select(f'chain {chain}')
        nt = selection.getResnums().tolist()
        pocketNT.append(np.unique(nt).tolist())
        
        # Export surface obj files for each pocket.
        cmd.load(f'{analysis}/{name}_out_real_sphere.pdb')
        cmd.hide('everything')
        cmd.remove(f'not resn STP or not resi {i+1}')
        cmd.alter('resn STP', 'vdw = b - 1.65')
        cmd.rebuild('all')
        cmd.set('surface_quality', 1)
        cmd.show('surface')
        cmd.save(
            f'{analysis}/pockets/pocket{i+1}_surf.obj',
            f'pocket_{i+1}_surface'
        )
        cmd.reinitialize()

        # Create mesh object for each pocket.
        try:
            mesh = trimesh.load(f'{analysis}/pockets/pocket{i+1}_surf.obj')
        except Exception as e:
            print(f"Error loading mesh for pocket {i+1}: {e}")
            pocket_npr1.append(np.nan)
            pocket_npr2.append(np.nan)
            continue

        # Calc inertia tensor for each pocket.
        try:
            inertia = mesh.moment_inertia
            if inertia.ndim != 2:
                raise ValueError("Inertia tensor is not 2-dimensional")
        except Exception as e:
            print(f"Error calculating inertia tensor for pocket {i+1}: {e}")
            pocket_npr1.append(np.nan)
            pocket_npr2.append(np.nan)
            continue

        # calculate eigen values and eigen vectors.
        try:
            eigvals, _ = np.linalg.eig(inertia)
        except Exception as e:
            print(f"Error calculating eigenvalues for pocket {i+1}: {e}")
            pocket_npr1.append(np.nan)
            pocket_npr2.append(np.nan)
            continue

        # sort eigen values > (I1, I2, I3)
        eig_ord = np.argsort(eigvals)

        # Calculate the normalize PMI ratios for each pocket.
        sorted_eigvals = eigvals[eig_ord]
        I1 = sorted_eigvals[0]
        I2 = sorted_eigvals[1]
        I3 = sorted_eigvals[2]
        npr1 = I1 / I3
        npr2 = I2 / I3
        pocket_npr1.append(npr1)
        pocket_npr2.append(npr2)

    # Check if lengths match
    if len(pocketNT) != len(pc_df.index):
        raise ValueError("Length of pocketNT does not match length of DataFrame index")

    # Add pocketNT and pocket npr data to pc dataframe.
    pc_df['PocketNT'] = pocketNT
    pc_df['Pocket_NPR1'] = pocket_npr1
    pc_df['Pocket_NPR2'] = pocket_npr2
    pc_df['Pocket_shape'] = 'Balanced'
    pc_df.loc[pc_df.eval('Pocket_NPR1 - Pocket_NPR2 + 0.5 < 0'), 'Pocket_shape'] = 'Rod-like'
    pc_df.loc[pc_df.eval('- Pocket_NPR1 - Pocket_NPR2 + 1.5 < 0'), 'Pocket_shape'] = 'Sphere-like'
    pc_df.loc[pc_df.eval('Pocket_NPR2 - 0.75 < 0'), 'Pocket_shape'] = 'Disc-like'

    # Add pocket filter (Pass or Fail) to pc dataframe.
    pc_df.loc[pc_df['Score'] > qualityfilter, 'Filter'] = 'Pass'

    # Check if pocketNT matches knownNT
    if isinstance(knownnt, list):
        pc_df['Type'] = pc_df.apply(
            lambda row: 'Known' if sum(nt in knownnt for nt in row['PocketNT']) >= 3 else row['Type'],
            axis=1
        )


def add_ligand_characteristics(
    analysis : str,
    stp_coords : prody.AtomGroup,
    ligand_coords : prody.AtomGroup,
    ligand : str,
    pc_df : pd.DataFrame,
) -> None:
    """Adds characteristics to the pocket characteristics DataFrame that
       require the presence of a ligand to calculate.
    Pocket overlap:  Ratio of a-spheres in contact with ligand.
    Ligand overlap:  Ratio of ligand atoms in contact with a-spheres.
    Center criteria: Distance from ligand to the geometric center of pocket.
    Type: Indicates if pocket is known or novel based on:
          pocket overlap, ligand overlap, and center criteria.

    Args:
        analysis (str): path directory contianing fpocket outputs for analysis.
        stp_coords (object): ProDy atom group of a-sphere coordinates.
        ligand_coords (object): ProDy atom group of ligand coordinates.
        ligand (str): Ligand residue name (usually a 3-letter code).
        pc_df (DataFrame): Characteristics and properities for each pocket.
    """

    pocket_overlap = []
    ligand_overlap = []
    center_criteria = []

    # Calculate the normalize PMI ratios for the ligand.
    ligand_npr1, ligand_npr2 = util.calc_npr(ligand_coords)

    #Calculate ligand QED score.
    if not os.path.isfile(f'{analysis}/{ligand}_ideal.sdf'):
        try:
            response = requests.get(
                f'https://www.ebi.ac.uk/pdbe/static/files/pdbechem_v2/{ligand}_ideal.sdf')
            with open(f'{analysis}/{ligand}_ideal.sdf', 'wb') as f:
                f.write(response.content)
        except:
            print(f'Error: Not able download ligand ideal structure for {ligand}.\n')
            qed = np.nan
    try:
        mol = Chem.MolFromMolFile(f'{analysis}/{ligand}_ideal.sdf')
        qed = QED.default(mol)
    except:
        print(f'Error: Not able calculate qed score for ligand {ligand}.\n')
        qed = np.nan

    # Iterates through each pocket (residue).
    for _, residue in enumerate(stp_coords.iterResidues()):
        pocket = residue.getCoords()

        # Calculate ratio of a-spheres in each pocket that is near the ligand.
        pocket_overlap_sele = residue.select(
            f'within 3 of ligand', ligand=ligand_coords)
        if pocket_overlap_sele is None:
            pocket_overlap.append(0)
        else:
            # Calculates total number of a-spheres in pocket.
            _, pocket_stp_count = np.unique(
                residue.getResnums(), return_counts=True)

            # Calculates number of overlapped a-spheres in pocket.
            _, pocket_overlap_stp_count = np.unique(
                pocket_overlap_sele.getResnums(), return_counts=True)

            pocket_overlap.append(float(
                pocket_overlap_stp_count / pocket_stp_count))

        # Calculate ratio of a-spheres in each pocket that is near the ligand.
        ligand_overlap_sele = ligand_coords.select(
            f'within 3 of pocket', pocket=residue)
        if ligand_overlap_sele is None:
            ligand_overlap.append(0)
        else:
            # Calculates total number of atoms in ligand.
            ligand_atom_count = ligand_coords.numAtoms()

            # Calculates number of overlapped atoms in ligand.
            ligand_overlap_atom_count = ligand_overlap_sele.numAtoms()

            ligand_overlap.append(
                ligand_overlap_atom_count / ligand_atom_count)
        
        # Calculate minimum distance between pocket center and ligand atoms.
        pocket_center = calcCenter(pocket)
        center_distance_l = calcDistance(pocket_center, ligand_coords)
        center_distance = min(center_distance_l)
        center_criteria.append(center_distance)
        
        
    pc_df['Ligand_ID'] = ligand
    # Add Ligand overlap and Center criteria to pc dataframe.
    pc_df['Pocket_overlap'] = pocket_overlap
    pc_df['Ligand_overlap'] = ligand_overlap
    pc_df['Center_criteria'] = center_criteria

    # Add pocket type (Known or Novel) to pc_df.
    pc_df.loc[
        (pc_df['Ligand_overlap'] >= 0.33) &
        (pc_df['Pocket_overlap'] >= 0.33) &
        (pc_df['Center_criteria'] <= 4), 'Type'
    ] = 'Known'

    # Add ligand NPR and QED score to pc_df.
    pc_df['QED_score'] = qed
    pc_df['Ligand_NPR1'] = ligand_npr1
    pc_df['Ligand_NPR2'] = ligand_npr2
    pc_df['Ligand_shape'] = 'Balanced'
    pc_df.loc[pc_df.eval('Ligand_NPR1 - Ligand_NPR2 + 0.5 < 0'), 'Ligand_shape'] = 'Rod-like'
    pc_df.loc[pc_df.eval('- Ligand_NPR1 - Ligand_NPR2 + 1.5 < 0'), 'Ligand_shape'] = 'Sphere-like'
    pc_df.loc[pc_df.eval('Ligand_NPR2 - 0.75 < 0'), 'Ligand_shape'] = 'Disc-like'



from rnavigate import plots, styles
import seaborn as sns


class Skyline(plots.Plot):
    """Plot per-nucleotide measurements as stepped line graphs (skyline plots).

    Parameters
    ----------
    num_samples : int
        Number of samples to plot.
    nt_length : int
        Length of the nucleotide sequence.
    region : tuple of int, defaults to "all" (entire sequence)
        start and end position of the region to plot. If "all", plot the entire
        sequence.
    **kwargs
        Keyword arguments passed to `rnavigate.plots.Plot`.

    Attributes
    ----------
    nt_length : int
        Length of the nucleotide sequence.
    region : tuple of int
        start and end position of the region to plot.
    track_height : float
        Height of the tracks in the plot.
    fig : matplotlib.figure.Figure
        Figure object.
    ax : matplotlib.axes.Axes
        Axes object.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    i : int
        Index of the current plot.
    """

    def __init__(self, num_samples, nt_length, region="all", **kwargs):
        """Initialize the plot."""
        if region == "all":
            self.nt_length = nt_length
            self.region = (1, nt_length)
        else:
            self.nt_length = region[1] - region[0] + 1
            self.region = region
        super().__init__(num_samples=num_samples, **kwargs)
        self.ax = self.axes[0, 0]
        self.track_height = 0

    def set_figure_size(
        self,
        height_ax_rel=None,
        width_ax_rel=0.03,
        width_ax_in=None,
        height_ax_in=2,
        height_gap_in=1,
        width_gap_in=0.5,
        top_in=1,
        bottom_in=1,
        left_in=1,
        right_in=1,
    ):
        """Set the figure size.

        Parameters
        ----------
        height_ax_rel : float
            Height of the axes relative to the y-axis limits.
        width_ax_rel : float
            Width of the axes relative to the x-axis limits.
        width_ax_in : float
            Width of the axes in inches.
        height_ax_in : float
            Height of the axes in inches.
        height_gap_in : float
            Height of the gap between axes in inches.
        width_gap_in : float
            Width of the gap between axes in inches.
        top_in : float
            Height of the top margin in inches.
        bottom_in : float
            Height of the bottom margin in inches.
        left_in : float
            Width of the left margin in inches.
        right_in : float
            Width of the right margin in inches.
        """

        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in + 2 * self.track_height,
            left_in=left_in,
            right_in=right_in,
        )

    def get_rows_columns(self, rows=None, cols=None):
        """Get the number of rows and columns.

        Parameters
        ----------
        rows : int
            Number of rows. This is ignored.
        cols : int
            Number of columns. This is ignored.

        Returns
        -------
        rows : int
            Number of rows. This is always 1.
        cols : int
            Number of columns. This is always 1.
        """
        return (1, 1)

    def get_ax(self, i=None):
        """Get the current axes object.

        Parameters
        ----------
        i : int, optional
            Index of the axes object. This is ignored.

        Returns
        -------
        ax : matplotlib.axes.Axes
            Axes object.
        """
        return self.ax

    def plot_data(
        self,
        profile,
        annotations=None,
        domains=None,
        label=None,
        columns=None,
        seqbar=True,
        errors=None,
        annotations_mode="track",
        nt_ticks=(20, 5),
    ):
        """Add data to the axes.

        Parameters
        ----------
        profile : rnavigate.profile.Profile
            Profile object.
        annotations : list of rnavigate.annotation.Annotation, optional
            List of annotation objects.
        domains : list of rnavigate.domain.Domain, optional
            List of domain objects.
        label : str, optional
            Sample name.
        columns : str or list of str, optional
            Which columns to plot. If None, plot the metric column.
        seqbar : bool, defaults to True
            Whether to plot a sequence bar.
        errors : str, optional
            Which error columns to plot. If None, do not plot errors.
        annotations_mode : "track" or "bar", defaults to "track"
            Whether to plot annotations as a track or as vertical bars.
        nt_ticks : tuple of int, defaults to (20, 5)
            Major and minor tick frequency for nucleotide positions.
        """
        if columns is None:
            columns = profile.metric
        track_unit = 0.03
        annotations_track = 2 * track_unit * len(annotations)
        annotations_track *= annotations_mode == "track"
        domains_track = 3 * track_unit * (domains is not None)
        sequence_track = track_unit * seqbar
        self.track_height = sequence_track + annotations_track + domains_track

        ax = self.get_ax()
        plots.plot_profile_skyline(ax, profile, label, columns, errors)
        if self.i == 0:
            if seqbar:
                plots.plot_sequence_track(
                    ax=ax,
                    sequence=profile.sequence,
                    yvalue=self.track_height * -1,
                    height=sequence_track,
                    region=self.region,
                    ytrans="axes",
                )
                self.add_colorbar_args(styles.get_nt_cmap())
            if domains is not None:
                for domain in domains:
                    plots.plot_domain_track(
                        ax=ax,
                        spans=domain,
                        yvalue=sequence_track - self.track_height,
                        height=domains_track,
                        region=self.region,
                        ytrans="axes",
                    )
            for i, annotation in enumerate(annotations):
                plots.plot_annotation_track(
                    ax=ax,
                    annotation=annotation,
                    mode=annotations_mode,
                    yvalue=(
                        sequence_track
                        + domains_track
                        - self.track_height
                        + (i + 0.5) * (2 * track_unit)
                    ),
                    height=2 * track_unit,
                    ytrans="axes",
                    region=self.region,
                )

        self.i += 1
        if self.i == self.length:
            if isinstance(columns, list):
                ylabel = [column.replace("_", " ") for column in columns]
                ylabel = ", ".join(ylabel)
            else:
                ylabel = columns.replace("_", " ")
            self.set_labels(ax=ax, ylabel=ylabel, axis_title=None, legend_title=None)
            self.set_axis(ax=ax, sequence=profile, nt_ticks=nt_ticks)

    def set_axis(self, ax, sequence, nt_ticks):
        """Set the axis limits and ticks.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object.
        sequence : rnavigate.data.Sequence
            The sequence on which position labels are based. Dashes are ignored.
        nt_ticks : tuple of int
            Major and minor tick frequency for nucleotide positions.
        """
        xlim = self.region
        ax.set_xlim([xlim[0] - 0.5, xlim[1] + 0.5])
        plots.set_nt_ticks(
            ax=ax,
            sequence=sequence,
            region=self.region,
            major=nt_ticks[0],
            minor=nt_ticks[1],
        )
        ax.spines["bottom"].set(
            position=("axes", self.track_height * -1), visible=False
        )
        ax.spines["left"].set_visible(False)
        ax.grid(axis="y", visible=True)
        ax.tick_params(axis="y", length=0, grid_alpha=0.4)
        sns.despine(ax=ax, bottom=True, left=True)

    def set_labels(
        self,
        ax,
        axis_title="Raw Reactivity Profile",
        legend_title="Samples",
        xlabel="Nucleotide Position",
        ylabel="Profile",
    ):
        """Set the axis labels and legend.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object.
        axis_title : str, optional
            Title of the axes.
        legend_title : str, optional
            Title of the legend.
        xlabel : str, optional
            Label of the x-axis.
        ylabel : str, optional
            Label of the y-axis.
        """
        ax.set_title(axis_title, loc="left")
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.legend(
            title=legend_title, labelcolor="linecolor", frameon=False, handlelength=0
        )



import numpy as np
from rnavigate import plots
from matplotlib.patches import Rectangle
from rnavigate.styles import rx_color, bg_color, dc_color


class SM(plots.Plot):
    """Plot classic ShapeMapper plots.

    Parameters
    ----------
    nt_length : int
        Length of the nucleotide sequence.
    region : tuple of int, defaults to "all" (entire sequence)
        start and end position of the region to plot. If "all", plot the entire
        sequence.
    panels : list of str, defaults to ["profile", "rates", "depth"]
        Which panels to plot. Options are "profile", "rates", and "depth".

    Attributes
    ----------
    nt_length : int
        Length of the nucleotide sequence.
    region : tuple of int
        start and end position of the region to plot.
    fig : matplotlib.figure.Figure
        Figure object.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    i : int
        Index of the current plot.
    """

    def __init__(self, nt_length, region=None, panels=["profile", "rates", "depth"]):
        """Initialize the plot."""
        if region is None:
            self.nt_length = nt_length
            self.region = [1, nt_length]
        super().__init__(len(panels), len(panels), cols=1)
        self.panels = panels

    def set_figure_size(
        self,
        height_ax_rel=None,
        width_ax_rel=0.03,
        width_ax_in=None,
        height_ax_in=2,
        height_gap_in=0.5,
        width_gap_in=0.5,
        top_in=1,
        bottom_in=0.5,
        left_in=0.5,
        right_in=0.5,
    ):
        """Set the figure size.

        Parameters
        ----------
        height_ax_rel : float, defaults to None
            Height of the axes relative to the y-axis limits.
        width_ax_rel : float, defaults to 0.03
            Width of the axes relative to the x-axis limits.
        width_ax_in : float, defaults to None
            Width of the axes in inches.
        height_ax_in : float, defaults to 2
            Height of the axes in inches.
        height_gap_in : float, defaults to 0.5
            Height of the gap between axes in inches.
        width_gap_in : float, defaults to 0.5
            Width of the gap between axes in inches.
        top_in : float, defaults to 1
            Height of the top margin in inches.
        bottom_in : float, defaults to 0.5
            Height of the bottom margin in inches.
        left_in : float, defaults to 0.5
            Width of the left margin in inches.
        right_in : float, defaults to 0.5
            Width of the right margin in inches.
        """
        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in,
            left_in=left_in,
            right_in=right_in,
        )

    def plot_data(self, profile, label):
        """Plot the data.

        Parameters
        ----------
        profile : rnavigate.profile.Profile
            Profile object.
        label : str
            Label for the plot.
        """
        self.fig.suptitle(label)
        for i, plot in enumerate(self.panels):
            ax = self.get_ax(i)
            plot_func = {
                "profile": self.plot_sm_profile,
                "rates": self.plot_sm_rates,
                "depth": self.plot_sm_depth,
            }[plot]
            plot_func(ax, profile)
        self.axes[-1, 0].set_xlabel("Nucleotide position")

    def plot_sm_profile(self, ax, profile):
        """Plots classic ShapeMapper profile on the given axis.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object.
        profile : rnavigate.profile.Profile
            Profile object.
        """
        ymin, ymax = (-0.5, 4)
        near_black = (0, 0, 1 / 255.0)
        orange_thresh = 0.4
        red_thresh = 0.85
        colors, _ = profile.get_colors("profile", profile=profile)
        sample = profile.data["Norm_profile"].copy()
        sample[np.isnan(sample)] = -1
        ax.bar(
            profile.data["Nucleotide"],
            sample,
            align="center",
            width=1.05,
            color=colors,
            edgecolor=colors,
            linewidth=0.0,
            yerr=profile.data["Norm_stderr"],
            ecolor=near_black,
            capsize=1,
        )
        ax.set_title("Normalized Profile")
        ax.set_ylim(ymin, ymax)
        ax.set_xlim(1, profile.length)
        ax.yaxis.grid(True)
        ax.set_axisbelow(True)
        ax.set_ylabel("Shape Reactivity")
        # add a SHAPE colorbar to the vertical ax
        # uses a little transformation magic to place correctly
        inv = ax.transData.inverted()
        for loc, spine in list(ax.spines.items()):
            if loc == "left":
                trans = spine.get_transform()
        tp = trans.transform_point([0, 0])
        tp2 = inv.transform_point(tp)
        rectX = tp2[0]
        tpA = (0, 0)
        tpB = (6, 0)
        tpA2 = inv.transform_point(tpA)
        tpB2 = inv.transform_point(tpB)
        rectW = tpB2[0] - tpA2[0]
        rect = Rectangle(
            (rectX, -0.5),
            rectW,
            orange_thresh + 0.5,
            facecolor="black",
            edgecolor="none",
        )
        ax.add_patch(rect)
        rect.set_clip_on(False)
        rect = Rectangle(
            (rectX, orange_thresh),
            rectW,
            red_thresh - orange_thresh,
            facecolor="orange",
            edgecolor="none",
        )
        ax.add_patch(rect)
        rect.set_clip_on(False)
        rect = Rectangle(
            (rectX, red_thresh),
            rectW,
            4 - red_thresh,
            facecolor="red",
            edgecolor="none",
        )
        ax.add_patch(rect)
        rect.set_clip_on(False)
        ax.get_xaxis().tick_bottom()  # remove unneeded ticks
        ax.get_yaxis().tick_left()
        ax.tick_params(axis="y", which="minor", left=False)
        ax.minorticks_on()
        ax.set(yticks=[0, 1, 2, 3, 4])

        for line in ax.get_yticklines():
            line.set_markersize(6)
            line.set_markeredgewidth(1)

        for line in ax.get_xticklines():
            line.set_markersize(7)
            line.set_markeredgewidth(2)

        for line in ax.xaxis.get_ticklines(minor=True):
            line.set_markersize(5)
            line.set_markeredgewidth(1)

        # put nuc sequence below ax
        plots.plot_sequence_track(ax, profile.sequence, yvalue=0, ytrans="axes")

    def plot_sm_depth(self, ax, profile):
        """Plots classic ShapeMapper read depth on the given axis.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object.
        profile : rnavigate.profile.Profile
            Profile object.
        """
        sample = profile.data
        ax.plot(
            sample["Nucleotide"],
            sample["Modified_read_depth"],
            linewidth=1.5,
            color=rx_color,
            alpha=1.0,
            label="Modified",
        )
        ax.plot(
            sample["Nucleotide"],
            sample["Untreated_read_depth"],
            linewidth=1.5,
            color=bg_color,
            alpha=1.0,
            label="Untreated",
        )
        ax.plot(
            sample["Nucleotide"],
            sample["Denatured_read_depth"],
            linewidth=1.5,
            color=dc_color,
            alpha=1.0,
            label="Denatured",
        )
        ax.set_xlim(1, profile.length)
        ax.legend(
            title="Effective depth\nin lighter colors",
            loc=2,
            borderpad=0.8,
            handletextpad=0.2,
            framealpha=0.75,
        )
        ax.plot(
            sample["Nucleotide"],
            sample["Modified_effective_depth"],
            linewidth=1.0,
            color=rx_color,
            alpha=0.3,
        )
        ax.plot(
            sample["Nucleotide"],
            sample["Untreated_effective_depth"],
            linewidth=1.0,
            color=bg_color,
            alpha=0.3,
        )
        ax.plot(
            sample["Nucleotide"],
            sample["Denatured_effective_depth"],
            linewidth=1.0,
            color=dc_color,
            alpha=0.3,
        )
        xmin, xmax, ymin, ymax = ax.axis()
        ax.set_ylim(0, ymax)
        ax.minorticks_on()
        ax.tick_params(axis="y", which="minor", left=False)
        yticks = [int(y) for y in ax.get_yticks()]
        formatted_ticks = [self.metric_abbreviate(val) for val in yticks]
        ax.set(yticks=yticks, yticklabels=formatted_ticks)
        for line in ax.get_yticklines():
            line.set_markersize(6)
            line.set_markeredgewidth(1)
        for line in ax.get_xticklines():
            line.set_markersize(7)
            line.set_markeredgewidth(2)
        for line in ax.xaxis.get_ticklines(minor=True):
            line.set_markersize(5)
            line.set_markeredgewidth(1)
        ax.yaxis.grid(True)
        ax.set_axisbelow(True)
        ax.set_ylabel("Read depth")
        # tried to make an offset, smaller font note about effective depths,
        # but couldn't get positioning/transforms to work properly.
        # For now just putting in xaxis label

    def plot_sm_rates(self, ax, profile):
        """Plots classic ShapeMapper mutation rates on the given axis.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object.
        profile : rnavigate.profile.Profile
            Profile object.
        """
        sample = profile.data
        # choose a decent range for ax, excluding high-background positions
        temp_rates = sample.loc[sample["Untreated_rate"] <= 0.05, "Modified_rate"]
        near_top_rate = np.nanpercentile(temp_rates, 98.0)
        maxes = np.array([0.32, 0.16, 0.08, 0.04, 0.02, 0.01])
        ymax = np.amin(maxes[maxes > near_top_rate])
        rx_err = sample["Modified_rate"] / sample["Modified_effective_depth"]
        rx_upper = sample["Modified_rate"] + rx_err
        rx_lower = sample["Modified_rate"] - rx_err
        bg_err = sample["Untreated_rate"] / sample["Untreated_effective_depth"]
        bg_upper = sample["Untreated_rate"] + bg_err
        bg_lower = sample["Untreated_rate"] - bg_err
        dc_err = sample["Denatured_rate"] / sample["Denatured_effective_depth"]
        dc_upper = sample["Denatured_rate"] + dc_err
        dc_lower = sample["Denatured_rate"] - dc_err
        ax.set_ylabel("Mutation rate (%)")
        ax.plot(
            sample["Nucleotide"],
            sample["Modified_rate"],
            zorder=3,
            color=rx_color,
            linewidth=1.5,
            label="Modified",
        )
        ax.plot(
            sample["Nucleotide"],
            sample["Untreated_rate"],
            zorder=2,
            color=bg_color,
            linewidth=1.5,
            label="Untreated",
        )
        ax.plot(
            sample["Nucleotide"],
            sample["Denatured_rate"],
            zorder=2,
            color=dc_color,
            linewidth=1.5,
        )
        ax.fill_between(
            sample["Nucleotide"],
            rx_lower,
            rx_upper,
            edgecolor="none",
            alpha=0.5,
            facecolor=rx_color,
        )
        ax.fill_between(
            sample["Nucleotide"],
            bg_lower,
            bg_upper,
            edgecolor="none",
            alpha=0.5,
            facecolor=bg_color,
        )
        ax.fill_between(
            sample["Nucleotide"],
            dc_lower,
            dc_upper,
            edgecolor="none",
            alpha=0.5,
            facecolor=dc_color,
        )
        ax.legend(loc=2, borderpad=0.8, handletextpad=0.2, framealpha=0.75)
        ax.set_xlim((1, len(sample["Modified_rate"])))
        ax.set_ylim((0, ymax))
        ax.minorticks_on()
        ax.tick_params(axis="y", which="minor", left=False)
        yticks = ax.get_yticks()
        yticklabels = [str(int(x * 100)) for x in yticks]
        ax.set(yticks=yticks, yticklabels=yticklabels)
        for line in ax.get_yticklines():
            line.set_markersize(6)
            line.set_markeredgewidth(1)
        for line in ax.get_xticklines():
            line.set_markersize(7)
            line.set_markeredgewidth(2)
        for line in ax.xaxis.get_ticklines(minor=True):
            line.set_markersize(5)
            line.set_markeredgewidth(1)
        ax.yaxis.grid(True)
        ax.set_axisbelow(True)

    def metric_abbreviate(self, num):
        """takes a large number and applies an appropriate abbreviation

        Parameters
        ----------
            num : int
                number to be abbreviated

        Returns
        -------
            str
                abbreviated number
        """
        suffixes = {3: "k", 6: "M", 9: "G"}
        s = str(num)
        # replace trailing zeros with metric abbreviation
        zero_count = len(s) - len(s.rstrip("0"))
        suffix = ""
        new_string = str(s)
        for num_zeros in sorted(suffixes.keys()):
            if num_zeros <= zero_count:
                suffix = suffixes[num_zeros]
                new_string = s[:-num_zeros]
                new_string = new_string + suffix
        return new_string



import pandas as pd
from rnavigate import data


class BedFile:
    def __init__(self, bedfile):
        self.bedfile = bedfile
        self.read_kwargs = {
            "header": None,
            "index_col": False,
            "usecols": [0, 1, 2, 3, 4, 5],
            "names": ["chr", "start", "end", "NA", "score", "strand"],
            "dtype": {
                "chr": str,
                "start": int,
                "end": int,
                "NA": str,
                "score": float,
                "strand": str,
            },
        }
        if self.bedfile.endswith(".gz"):
            self.read_kwargs["compression"] = "gzip"
        self.profile_cols = ["Score"]

    def get_annotation(self, transcript, **kwargs):
        return self.get_annotations([transcript], **kwargs)[transcript]

    def get_annotations(self, transcripts, **kwargs):
        bed_df = pd.read_table(self.bedfile, **self.read_kwargs)
        bed_df["start"] += 1  # bed files are 0-indexed, closed right
        tx_annotations = {}
        for tx in transcripts:
            chrom = tx.chromosome
            strand = tx.strand
            tx_coordinates = tx.coordinate_df["Coordinate"]
            mn = tx_coordinates.min()
            mx = tx_coordinates.max()
            if strand == "-":
                mn, mx = mx, mn
            tx_df = bed_df.query(
                f'(chr == "{chrom}") '
                f'& (strand == "{strand}") '
                f"& ((start > {mn} and start < {mx})"
                f"| (end > {mn} and end < {mx}))"
            )
            spans = []
            for _, row in tx_df.iterrows():
                span = tx.get_tx_range(row["start"], row["end"])
                if span[0] != span[1]:
                    spans.append(span)
            tx_annotations |= {
                tx: data.Annotation(
                    input_data=spans,
                    annotation_type="spans",
                    sequence=tx.sequence,
                    **kwargs,
                )
            }
        return tx_annotations

    def get_profile(self, transcript, **kwargs):
        bed_df = pd.read_table(self.bedfile, **self.read_kwargs)
        chrom = transcript.chromosome
        strand = transcript.strand
        profile = transcript.coordinate_df.copy()
        mn = profile["Coordinate"].min()
        mx = profile["Coordinate"].max()
        bed_df = bed_df.query(
            f'chr == "{chrom}" & strand == "{strand}" & '
            f"((start > {mn} & start < {mx}) | (end > {mn} & end < {mx}))"
        )
        profile[self.profile_cols] = 0
        for _, row in bed_df.iterrows():
            site = profile.eval(
                f'Coordinate > {row["start"]} and Coordinate < {row["end"]}'
            )
            for col in self.profile_cols:
                profile.loc[site, col] = row[col]
        return data.Profile(input_data=profile, metric="Value", **kwargs)

    def get_density_profile(self, transcript, **kwargs):
        bed_df = pd.read_table(self.bedfile, **self.read_kwargs)
        chrom = transcript.chromosome
        strand = transcript.strand
        profile = transcript.coordinate_df.copy()
        mn = profile["Coordinate"].min()
        mx = profile["Coordinate"].max()
        bed_df = bed_df.query(
            f'chr == "{chrom}" & strand == "{strand}" & '
            f"((start > {mn} & start < {mx}) | (end > {mn} & end < {mx}))"
        )
        profile["Density"] = 0
        for _, row in bed_df.iterrows():
            site = profile.eval(
                f'Coordinate > {row["start"]} and Coordinate < {row["end"]}'
            )
            profile.loc[site, "Density"] += 1
            for col in self.profile_cols:
                profile.loc[site, col] += row[col]
        return data.Profile(input_data=profile, metric="Value", **kwargs)


class NarrowPeak(BedFile):
    def __init__(self, bedfile):
        super().__init__(bedfile=bedfile)
        self.read_kwargs["usecols"].extend([6, 7, 8, 9])
        self.read_kwargs["names"].extend(["Value", "P_value", "Q_value", "Peak"])
        self.read_kwargs["dtype"].update(
            {"Value": float, "P_value": float, "Q_value": float, "Peak": int}
        )
        self.profile_cols = ["Score", "Value", "P_value", "Q_value", "Peak"]



#!/usr/bin/env python3
# -----------------------------------------------------
# pymol figure creation code
# Seth Veenbaas
# Weeks Lab, UNC-CH
# 2022
#
# Version 1.3.0
#
# -----------------------------------------------------
import os
from pymol import util
from pymol import cmd


def load_pdb(pdb: str) -> None:
    """load the input pdb file into a pymol session

    Args:
        pdb (str): path to pdb file
    """
    cmd.load(f'{pdb}', partial=1)


def alignligand(mobile: str, target: str, target_state: int = 0) -> None:
    """aligns mobile structure to target structure using PyMol align

    Args:
        mobile (str): Path mobile .pdb file.
        target (str): Path to target .pdb file.
    """
    target_object = os.path.splitext(os.path.basename(target))[0]
    objects_list = cmd.get_object_list()
    loaded = target_object in objects_list
    if not loaded:
        cmd.load(target, object=target_object)
        
    if target_state is None or \
        target_state > cmd.count_states(target_object) or \
        cmd.count_states(target_object) == 1:
        target_state = 0

    # aligns mobile object to target object
    cmd.align(mobile, target_object, cycles=10, target_state=target_state)
    cmd.set('cartoon_color', 'palecyan', target_object)
    cmd.disable(f'{target_object}')


def set_default() -> None:
    """Set default pymol settings in current pymol session:
    -raytracing performance
    -structure style

    Settings are from the pymolrc.pml file at:
    https://github.com/Weeks-UNC/small-scripts/tree/master/Pymol
    """
    # enable multi-thread processing
    cmd.set('max_threads', 16)

    # increase raytracing memory allowance
    cmd.set('hash_max', 2048)

    # change backround color
    cmd.bg_color('white')

    # sets specular reflection intensity
    cmd.set('specular', '0.1')

    # controls appearence of shadows for ray-traced images
    cmd.set('ray_shadows', 'off')

    # controls antiliasing/edge smoothing for ray-traced images
    cmd.set('antialias', '2')

    # orthoscopic turns on and off the perspective handling
    cmd.set('orthoscopic', 'off')

    # set trasparent background
    cmd.set('ray_opaque_background', '0')

    # raytraces full color without outlines
    cmd.set('ray_trace_mode', '0')

    # settings related to surface features
    cmd.set('surface_quality', '1')
    cmd.set('solvent_radius', '1.6')
    cmd.set('transparency', '0.6')
    cmd.set('surface_color', 'grey80')

    # settings related to rendering meshes
    cmd.set('mesh_quality', '2')
    cmd.set('mesh_type', '0')
    cmd.set('mesh_width', '0.5')
    cmd.set('mesh_radius', '0.015')

    # RNA style settings
    cmd.hide('sticks', 'polymer')
    cmd.set('cartoon_ring_mode', '3')
    cmd.set('cartoon_ring_finder', '1')
    cmd.remove('resn hoh')
    cmd.remove('inorganic and not resn STP')
    cmd.cartoon('oval', 'polymer')
    cmd.set('cartoon_oval_length', '0.75')
    cmd.set('cartoon_oval_width', '0.25')
    cmd.set_color('greyish', [0.625, 0.7, 0.7])
    cmd.set_color('novel', [0.0, 0.4314322351168238, 0.1118361643280874])
    cmd.set_color('known', [0.908605075491407, 0.3955005147576708, 0.0])
    cmd.color('greyish', 'polymer')
    util.cbao('organic')
    cmd.color('lightpink', '(byres polymer & name CA)')
    cmd.cartoon('automatic', '(byres polymer & name CA)')


def color_pockets(
        pocket_cmap : dict[int, tuple[float, float, float, float]]
    ) -> None:
    """settings appearance of the a-spheres (STP) that compose pockets:
    -sets a-sphere radius
    -colors a-spheres based on input color map

    Args:
        pocket_cmap (Dict[int,tuple(int,int,int)]): 
            key: pocket index
            value: rgb color value (tuple)
    """
    cmd.hide('everything', 'resn STP')

    # Alters sphere radius to actual size of a-core.
    cmd.alter('resn STP', 'vdw = b - 1.65')

    # colors each pocket based on color map
    for pocket_num in pocket_cmap:
        cmd.show('spheres', f'resn STP and resi {str(pocket_num)}')
        rgb_tuple = pocket_cmap[pocket_num]
        rgb_list = list(rgb_tuple[0:3])
        cmd.set_color(f'pocket{str(pocket_num)}_color', rgb_list)
        cmd.color(f'pocket{str(pocket_num)}_color',
                  f'resn STP and resi {str(pocket_num)}')


def color_multistate_pockets(
        object_name : str,
        pocket_cmap : dict[int, tuple[float, float, float, float]],
    ) -> None:
    cmd.extract(f'{object_name}_pockets', f'{object_name} and resn STP')
    cmd.alter(f'{object_name}_pockets', 'vdw = b - 1.65')
    for pocket_num in pocket_cmap:
        rgb_tuple = pocket_cmap[pocket_num]
        rgb_list = list(rgb_tuple[0:3])
        cmd.set_color(f'{object_name}_pocket{str(pocket_num)}_color', rgb_list)
        cmd.color(
            f'{object_name}_pocket{str(pocket_num)}_color',
            f'{object_name}_pockets and resi {str(pocket_num)}'
        )
        cmd.set(
            'surface_color',
            f'{object_name}_pocket{str(pocket_num)}_color', 
            f'{object_name}_pockets and resi {str(pocket_num)}'
        )


def transparent_pocket(
        object_name : str,
        state : int,
        multistate_pocket_cmap : dict[int, tuple[float,float,float,float]]
    ) -> None:
    cmd.hide('sticks', f'{object_name}_pockets')
    cmd.hide('spheres', f'{object_name}_pockets')
    show_pocket_resi = list(multistate_pocket_cmap[state].keys())
    for resi in show_pocket_resi:
        cmd.show('surface', f'{object_name}_pockets and resi {resi}')
    cmd.set('transparency', '0.80', f'{object_name}_pockets')


def make_multistate() -> None:
    cmd.set('cartoon_ring_finder', '0')
    cmd.set('cartoon_ring_mode', '1')
    cmd.set('cartoon_transparency', '0.90')
    cmd.set('transparency_mode', '3')


def save_3D_figure(
    path: str, name: str, dpi: int, chain: str, zoom: int
    ) -> None:
    """Saves pymol session file and raytraced png file.

    Args:
        analysis (str): path directory contianing fpocket outputs for analysis
        name (str): name of output pdb structure
        dpi (int): Figure resolution in dpi (dots per linear inch).
        c (str): chain of anzylzed structure
        zoom (float): Sets zoom buffer distance (Å) for creating 3D figures.
    """

    if ',' in chain:
        chain = chain.replace(',', '+')
    cmd.remove('hydrogens')
    cmd.set('ray_trace_fog', '0')
    cmd.orient()
    # cmd.rotate('z', '90')
    cmd.zoom(f'chain {chain} and (byres polymer & name O2)', f'{zoom}', complete=1)
    cmd.save(f'{path}/{name}_out_real_sphere.pse')
    dimension = dpi * 8
    print(f'Ray tracing: {name}...')
    cmd.png(f'{path}/{name}_3D_{dpi}.png',
            width=dimension, height=dimension, dpi=dpi, ray=1)
    cmd.reinitialize()



"""Module for plotting 3Dmol.js viewers"""

import py3Dmol
import matplotlib.colors as mpc
from rnavigate import plots


class Mol(plots.Plot):
    def __init__(
        self,
        num_samples,
        pdb,
        width=400,
        height=400,
        background_alpha=1,
        rotation=None,
        orientation=None,
        style="cartoon",
        rows=None,
        cols=None,
    ):
        """Create a 3Dmol.js viewer with a grid of subviewers

        Parameters
        ----------
        num_samples : int
            Number of subviewers to create
        pdb : rnavigate.pdb.PDB
            PDB object to use for plotting
        width : int, defaults to 400
            Width of each subviewer in pixels
        height : int, defaults to 400
            Height of each subviewer in pixels
        background_alpha : float, defaults to 1
            Alpha value for the background color
        rotation : dict, defaults to None
            Dictionary of rotation angles for the viewer
        orientation : list of floats
            List of floats defining the orientation of the viewer
        style : str, defaults to "cartoon"
            Style of the viewer
        rows : int, defaults to None
            Number of rows to use in the grid of subviewers. If None, the
            number of rows is determined automatically.
        cols : int, defaults to None
            Number of columns to use in the grid of subviewers. If None, the
            number of columns is determined automatically.

        Attributes
        ----------
        view : py3Dmol.view
            3Dmol.js viewer object
        i : int
            Index of the current subviewer
        rows : int
            Number of rows in the grid of subviewers
        columns : int
            Number of columns in the grid of subviewers
        colorbars : list of matplotlib.colorbar.ColorbarBase
            List of colorbars to be added to the plot
        style : str
            Style of the viewer
        pdb : rnavigate.pdb.PDB
            PDB object to use for plotting
        length : int
            Number of samples
        """
        self.colorbars = []
        self.style = style
        self.pdb = pdb
        self.length = num_samples
        self.rows, self.columns = self.get_rows_columns(rows=rows, cols=cols)
        view = py3Dmol.view(
            viewergrid=(self.rows, self.columns),
            width=width * self.columns,
            height=height * self.rows,
        )
        view.setBackgroundColor("white", background_alpha)
        with open(self.pdb.path, "r") as pdb_file:
            pdb_str = pdb_file.read()
        if self.pdb.path.split(".")[-1] == "pdb":
            view.addModel(pdb_str, "pdb")
        elif self.pdb.path.split(".")[-1] == "cif":
            view.addModel(pdb_str, "mmcif")
        view.setStyle({self.style: {"color": "spectrum"}})
        view.zoomTo({"chain": self.pdb.chain})
        if orientation is not None:
            view.setView(orientation)
        elif rotation is not None:
            for key in rotation:
                view.rotate(rotation[key], key)
        self.view = view
        self.i = 0

    def get_viewer(self, i=None):
        """Get the subviewer at index i"""
        if i is None:
            i = self.i
        row = i // self.columns
        col = i % self.columns
        return (row, col)

    def plot_data(
        self,
        interactions=None,
        profile=None,
        label=None,
        colors="grey",
        atom="O2'",
        title=True,
        get_orientation=False,
        viewer=None,
    ):
        """Plot data on the current subviewer.

        Parameters
        ----------
        interactions : rnavigate.interactions.Interactions
            Interactions object to plot as lines on the 3d structure
        profile : rnavigate.profile.Profile
            Profile object to use as nucleotide colors
        label : str, defaults to None
            Label to use as a title on the subviewer
        colors : str, defaults to "grey"
            Color scheme to use for nucleotides
        atom : str, defaults to "O2'"
            Atom to use for plotting interactions
        title : bool, defaults to True
            Whether or not to add a title to the subviewer
        get_orientation : bool, defaults to False
            Whether or not to get the orientation of the subviewer
            This will display the orientation as a label on the subviewer when the
            structure is clicked.
        viewer : tuple of ints, defaults to None
            Tuple of ints defining the subviewer to plot on
        """
        if viewer is None:
            viewer = self.get_viewer()
        if get_orientation:
            self.get_orientation()
        if interactions is not None:
            self.plot_interactions(viewer, interactions, atom)
        self.set_colors(viewer, profile, colors)
        if title:
            self.view.addLabel(
                label,
                {
                    "position": {"x": 0, "y": 0, "z": 0},
                    "useScreen": True,
                    "fontColor": "black",
                    "backgroundOpacity": 0.0,
                    "fontSize": 28,
                },
                viewer=viewer,
            )
        self.i += 1

    def add_lines(self, i, j, color, viewer, atom):
        """Add lines between nucleotides i and j

        Parameters
        ----------
        i : int
            Index of the first nucleotide
        j : int
            Index of the second nucleotide
        color : str
            Color to use for the line
        viewer : tuple of ints
            Tuple of ints defining the subviewer to plot on
        atom : str
            Atom to use for plotting interactions
        """
        pdb = self.pdb
        if not pdb.is_valid_idx(seq_idx=i) or not pdb.is_valid_idx(seq_idx=j):
            return
        xi, yi, zi = pdb.get_xyz_coord(i, atom)
        xj, yj, zj = pdb.get_xyz_coord(j, atom)
        cylinder_specs = {
            "start": {"x": xi, "y": yi, "z": zi},
            "end": {"x": xj, "y": yj, "z": zj},
            "radius": 0.5,
            "fromCap": 2,
            "toCap": 2,
            "color": color,
        }
        self.view.addCylinder(cylinder_specs, viewer=viewer)

    def plot_interactions(self, viewer, interactions, atom):
        """Plot interactions on the current subviewer

        Parameters
        ----------
        viewer : tuple of ints
            Tuple of ints defining the subviewer to plot on
        interactions : rnavigate.interactions.Interactions
            Interactions object to plot as lines on the 3d structure
        atom : str
            Atom to use for plotting interactions
        """
        window = interactions.window
        for i, j, color in zip(*interactions.get_ij_colors()):
            color = "0x" + mpc.rgb2hex(color)[1:]
            for w in range(window):
                io = i + w
                jo = j + window - 1 - w
                self.add_lines(io, jo, color, viewer, atom)
        self.add_colorbar_args(interactions.cmap)

    def set_colors(self, viewer, profile, colors):
        """Set the colors of the nucleotides on the current subviewer

        Parameters
        ----------
        viewer : tuple of ints
            Tuple of ints defining the subviewer to plot on
        profile : rnavigate.profile.Profile
            Profile object to use as nucleotide colors
        colors : str
            Color scheme to use for nucleotides
        """
        colors, _ = self.pdb.get_colors(colors, profile=profile)
        color_selector = {}
        valid_pdbres = []
        for res in self.pdb.pdb_idx:
            res = int(res)
            valid_pdbres.append(res)
            color = colors[self.pdb.get_seq_idx(res) - 1]
            if color in color_selector:
                color_selector[color].append(res)
            else:
                color_selector[color] = [res]
        for color, selector in color_selector.items():
            selector = {"chain": self.pdb.chain, "resi": selector}
            style = {self.style: {"color": color}}
            self.view.setStyle(selector, style, viewer=viewer)
        selector = {"chain": self.pdb.chain, "resi": valid_pdbres, "invert": "true"}
        style = {"cross": {"hidden": "true"}}
        self.view.setStyle(selector, style, viewer=viewer)

    def hide_cylinders(self):
        """Hide the cylinders that represent nucleotides."""
        resns = ["A", "G", "C", "U"]
        atoms = ["N1", "N1", "N3", "N3"]
        for resn, atom in zip(resns, atoms):
            self.view.setStyle(
                {"resn": resn, "atom": atom}, {"cross": {"hidden": "true"}}
            )

    def save(self):
        """Display the current orientation of the viewer as a png image.

        Notes
        -----
        This method must be run in a new cell after the viewer has been
        instantiated. The resulting png image will be saveable as a png file by
        clicking and dragging the image to your desktop.
        """
        print(
            "To save, orient the interactive plot to the view you'd like\n"
            "save, then run plot.save() in a new cell. The resulting image\n"
            "will be saveable as a png file"
        )
        return self.view.png()

    def get_orientation(self):
        """Adds a clickable event to the viewer to display the orientation vector."""
        self.view.setClickable(
            {},
            "true",
            """function(atom,viewer,event,container){
                viewer.addLabel(viewer.getView().map(function(each_element){return each_element.toFixed(2)}),
                {
                    position: {x:0,y:0,z:0},
                    useScreen: true,
                    backgroundColor: 'black'
                });
            }""",
        )



"""The PDB object to represent tertiary structures with atomic coordinates.

This data can be used to filter interactions by 3D distance, and to visualize
profile and interactions data on interactive 3D structures.
"""

import Bio.PDB
from rnavigate import data
import numpy as np


class PDB(data.Sequence):
    """A class to represent RNA tertiary structures with atomic coordinates.

    This data can be used to filter interactions by 3D distance, and to visualize
    profile and interactions data on interactive 3D structures.

    Parameters
    ----------
    input_data : str
        path to a PDB or CIF file
    chain : str
        chain identifier of RNA of interest
    sequence : rnavigate.Sequence or str, optional
        A sequence to use as the reference sequence.
        This is required if the sequence cannot be found in the header
        Defaults to None.
    name : str, optional
        A name for the data set. Defaults to None.

    Attributes
    ----------
    sequence : str
        The RNA sequence
    length : int
        The length of the RNA sequence
    name : str
        A name for the data set
    path : str
        The path to the PDB or CIF file
    chain : str
        The chain identifier of the RNA of interest
    offset : int
        The offset between the sequence positions and the PDB residue indices
    pdb : Bio.PDB.Structure.Structure
        The PDB structure
    pdb_idx : np.array
        The PDB indices of the RNA
    pdb_seq : np.array
        The PDB sequence of the RNA
    distance_matrix : dict
        A dictionary of distance matrices for each atom type
    """

    def __init__(self, input_data, chain, sequence=None, name=None):
        """Construct PDB object based on an input PDB or CIF file."""
        self.offset = 0
        self.chain = chain
        if sequence is None:
            sequence = self.get_sequence(input_data)
        super().__init__(sequence, name=name)
        self.read_pdb(input_data)
        self.path = input_data
        self.distance_matrix = {}

    def get_sequence(self, pdb):
        """Find the sequence in the provided CIF or PDB file.

        Parameters
        ----------
        pdb : str
            path to a PDB or CIF file

        Returns
        -------
        sequence : string
            The RNA sequence
        """
        seqres = []
        with open(pdb) as file:
            for line in file.readlines():
                line = [field.strip() for field in line.split()]
                if line[0] == "SEQRES" and line[2] == self.chain:
                    seqres.extend(line[4:])
        return self.get_sequence_from_seqres(seqres)

    def get_sequence_from_seqres(self, seqres):
        """Used by get_sequence to parse the SEQRES entries.

        Parameters
        ----------
        seqres : list
            A list of SEQRES entries for the RNA chain of interest

        Returns
        -------
        sequence : string
            The RNA sequence
        """
        sequence = []
        for nt in seqres:
            valid = nt[0].upper() in "GUACT"
            if valid and (len(nt) == 3) and nt.endswith("TP"):
                sequence.append(nt[0])
            elif valid and (len(nt) == 1):
                sequence.append(nt[0])
            else:
                raise ValueError("invalid nt in SEQRES: " + nt)
        return "".join(sequence)

    def read_pdb(self, pdb):
        """Read a PDB or CIF file into the data structure.

        Parameters
        ----------
        pdb : str
            path to a PDB or CIF file
        """
        if pdb.split(".")[-1] == "pdb":
            parser = Bio.PDB.PDBParser(QUIET=True)
        elif pdb.split(".")[-1] == "cif":
            parser = Bio.PDB.MMCIFParser(QUIET=True)
        self.pdb = parser.get_structure("RNA", pdb)
        self.pdb_idx = []
        self.pdb_seq = []
        for res in self.pdb[0][self.chain].get_residues():
            res_id = res.get_id()
            res_seq = res.get_resname()
            if res_id[0] == " ":
                self.pdb_idx.append(res_id[1])
                self.pdb_seq.append(res_seq.strip())
        self.pdb_idx = np.array(self.pdb_idx)
        self.set_indices()

    def set_indices(self):
        """Uses self.data and self.sequence to set self.offset"""
        for i in range(len(self.sequence)):
            correct_offset = True
            for pdb_nt, pdb_idx in zip(self.pdb_seq, self.pdb_idx):
                if self.sequence[pdb_idx - i - 1] != pdb_nt:
                    correct_offset = False
                    break
            if correct_offset:
                self.offset = i
                break
        if not correct_offset:
            print("PDB entries could not be matched to sequence.")

    def get_pdb_idx(self, seq_idx):
        """Return the PDB index given the sequence index (0-indexed)."""
        return seq_idx + self.offset

    def get_seq_idx(self, pdb_idx):
        """Return the sequence index given the PDB index."""
        return pdb_idx - self.offset

    def is_valid_idx(self, pdb_idx=None, seq_idx=None):
        """Determines if a PDB or sequence index is in the PDB structure.

        Parameters
        ----------
        pdb_idx : int, optional
            A PDB index (1-indexed). Defaults to None.
        seq_idx : int, optional
            A sequence index (1-indexed). Defaults to None.

        Returns
        -------
        bool
            True if the index is in the PDB structure, False otherwise.
        """
        if pdb_idx is not None and pdb_idx in self.pdb_idx:
            return True
        elif seq_idx is not None and (seq_idx in (self.pdb_idx - self.offset)):
            return True
        else:
            return False

    def get_xyz_coord(self, nt, atom):
        """Return the x, y, and z coordinates for a given residue and atom.

        Parameters
        ----------
        nt : int
            The nucleotide of interest (1-indexed)
        atom : string or dict, defaults to "O2'"
            The atom to use for distance calculations. If a string, the same atom
            will be used for all residues. If a dict, the atom will be chosen based
            on the nucleotide type. If "DMS", the N1 atom will be used for A and G,
            and the N3 atom will be used for U and C.

        Returns
        -------
        xyz : list
            A list of x, y, and z coordinates
        """
        pdb_idx = self.get_pdb_idx(nt)
        if atom == "DMS":
            atom = {nt: "N1" for nt in "AG"} | {nt: "N3" for nt in "UC"}
        elif isinstance(atom, str):
            atom = {nt: atom for nt in "AUCG"}
        seq = self.sequence[nt - 1].upper().replace("T", "U")
        atom = atom[seq]
        xyz = [
            float(c) for c in self.pdb[0][self.chain][int(pdb_idx)][atom].get_coord()
        ]
        return xyz

    def get_distance(self, i, j, atom="O2'"):
        """Get the distance between given atom in nucleotides i and j (1-indexed).

        Parameters
        ----------
        i : int
            The first nucleotide
        j : int
            The second nucleotide
        atom : string or dict, defaults to "O2'"
            The atom to use for distance calculations. If a string, the same atom
            will be used for all residues. If a dict, the atom will be chosen based
            on the nucleotide type. If "DMS", the N1 atom will be used for A and G,
            and the N3 atom will be used for U and C.

        Returns
        -------
        distance : float
            The distance between the atoms
        """
        if atom in self.distance_matrix:
            return self.distance_matrix[atom][i - 1, j - 1]
        try:
            xi, yi, zi = self.get_xyz_coord(i, atom)
            xj, yj, zj = self.get_xyz_coord(j, atom)
            distance = ((xi - xj) ** 2 + (yi - yj) ** 2 + (zi - zj) ** 2) ** 0.5
        except KeyError:
            distance = np.nan
        return distance

    def get_distance_matrix(self, atom="O2'"):
        """Get the pairwise atomic distance matrix for all residues.

        Parameters
        ----------
        atom : string or dict, defaults to "O2'"
            The atom to use for distance calculations. If a string, the same atom
            will be used for all residues. If a dict, the atom will be chosen based
            on the nucleotide type. If "DMS", the N1 atom will be used for A and G,
            and the N3 atom will be used for U and C.

        Returns
        -------
        matrix : NxN numpy.ndarray
            A 2D array of pairwise distances. N is the length of the RNA.
        """
        if atom in self.distance_matrix:
            return self.distance_matrix[atom]
        x = np.full(self.length, np.nan)
        y = np.full(self.length, np.nan)
        z = np.full(self.length, np.nan)
        for i in self.pdb_idx - self.offset:
            try:
                x[i - 1], y[i - 1], z[i - 1] = self.get_xyz_coord(i, atom)
            except KeyError:
                pass
        a = x - x[:, np.newaxis]
        b = y - y[:, np.newaxis]
        c = z - z[:, np.newaxis]
        matrix = np.sqrt(a * a + b * b + c * c)
        # self.distance_matrix[atom] = np.nan_to_num(matrix, nan=1000)
        return matrix
